/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: B.c
  function: Ogg machine level bitpacking/backend routines
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Aug 01 1996

 ********************************************************************/

#include "ogg.h"
#include "memory.h"
#include "bitwise.h"
#include "parser.h"
#include "virtual.h"
#include "ilog.h"
#include "huffman.h"
#include "B.h"

extern int OM_MACHINE;

/* Arithmetic-like hasn't been ported from 0.97 yet */
/* VQ will go here once we sit down and figure out what clustering
   algs we want to use.  For now, huff is sufficient */

/* Removed write-to-write state handling code as it seems unlikely
   that arithmetic will ever be used from 0.98 on, and there are no
   other encoders with inter-write state */

/*****************************************************************/
/* LOW-LEVEL function registry                                   */
/*****************************************************************/

/* the higher level encoder struct holds our dictionary information;
   these are merely hooks for functions */

static void (*lowB_undo(module *m))(int){
  if(!strcmp(m->name,"Bh"))
    return(__ogg_Hundodict);

  __ogg_Cerror(104,0); /* Gee, must not have been a compatible function */
  return(NULL); /* Avoid the warning */
}

static void (*lowB_commit(module *m))(int){
  if(!strcmp(m->name,"Bh"))
    return(__ogg_Hcheckpointdict);

  __ogg_Cerror(104,0); /* Gee, must not have been a compatible function */
  return(NULL); /* Avoid the warning */
}

static int (*lowB_create(module *m))(int,int,int,int){
  if(!strcmp(m->name,"Bh"))
    return(__ogg_Hcreatedict);

  __ogg_Cerror(104,0); /* Gee, must not have been a compatible function */
  return(NULL); /* Avoid the warning */
}

static void (*lowB_write(module *m))(int,int,int,int){
  if(!strcmp(m->name,"Bh"))
    return(__ogg_Hwrite);

  __ogg_Cerror(104,0); /* Gee, must not have been a compatible function */
  return(NULL); /* Avoid the warning */
}

static int (*lowB_read(module *m))(int,int){
  if(!strcmp(m->name,"Bh"))
    return(__ogg_Hread);

  __ogg_Cerror(104,0); /* Gee, must not have been a compatible function */
  return(NULL); /* Avoid the warning */
}

static void (*lowB_softreset(module *m))(int){
  if(!strcmp(m->name,"Bh"))
    return(__ogg_Hresetdict);

  __ogg_Cerror(104,0); /* Gee, must not have been a compatible function */
  return(NULL); /* Avoid the warning */
}

long __ogg_Bmax;
long __ogg_Bbits;
void *__ogg_Bstate; /* The dict writing functions need the dict state */

/******** B ******************************************************/

/* creation, soft-reset,etc are nops */

static void B_write(long val){
  if(__ogg_Bbits)
    __ogg_Bout((unsigned long)val,__ogg_Bbits);
  else
    __ogg_Bout((unsigned long)val,ilog(__ogg_Bmax));
}

static long B_read(void){
  if(__ogg_Bbits)
    return(__ogg_Bread(__ogg_Bbits));
  else
    return(__ogg_Bread(ilog(__ogg_Bmax)));
}

/******** Bdpl ***************************************************/

typedef struct Bdpl{
  int *handles;
  int bits;

  module *backend;
  void (*write)(int,int,int,int);
  int (*read)(int,int);
}Bdpl;

static void Bdpl_write(long val){
  int max,maxbits;
  Bdpl *b=(Bdpl *)__ogg_Bstate;
  if(__ogg_Bbits){
    maxbits=__ogg_Bbits;
    max=(1<<maxbits)-1;
  }else{
    max=__ogg_Bmax;
    maxbits=ilog(max);
  }

  if(maxbits>b->bits){
    int shift=maxbits-b->bits;
    (*(b->write))(b->handles[max>>shift],val>>shift,1,1);
    __ogg_Bout(val,shift);
  }else
    (*(b->write))(b->handles[max],val,1,1);
}

static long Bdpl_read(void){
  int max,maxbits,val;
  Bdpl *b=(Bdpl *)__ogg_Bstate;
  if(__ogg_Bbits){
    maxbits=__ogg_Bbits;
    max=(1<<maxbits)-1;
  }else{
    max=__ogg_Bmax;
    maxbits=ilog(max);
  }

  if(maxbits>b->bits){
    int shift=maxbits-b->bits;
    val=(*(b->read))(b->handles[max>>shift],1);
    val=(val<<shift) | __ogg_Bread(shift);
  }else
    val=(*(b->read))(b->handles[max],1);

  return(val);
}

static void Bdpl_softreset(module *m){
  int i;
  Bdpl *b=(Bdpl *)m->state;

  for(i=1;i<(1<<(b->bits));i++)
    (*(lowB_softreset(b->backend)))(b->handles[i]);
}

static void Bdpl_undo(void){
  int i;
  Bdpl *b=(Bdpl *)__ogg_Bstate;

  for(i=1;i<(1<<(b->bits));i++)
    (*(lowB_undo(b->backend)))(b->handles[i]);
}

static void Bdpl_commit(void){
  int i;
  Bdpl *b=(Bdpl *)__ogg_Bstate;

  for(i=1;i<(1<<(b->bits));i++)
    (*(lowB_commit(b->backend)))(b->handles[i]);
}

void __oggc_Bdpl(module *m){
  Bdpl *b=(Bdpl *)(m->state=__ogg_calloc(1,sizeof(Bdpl),OM_MACHINE));
  int i;
  b->bits=__ogg_fixedARG(m,0);
  b->backend=__ogg_gethandleM(__ogg_fixedARG(m,1));
  if(__ogg_fixedARG(m,4))
    b->read=lowB_read(b->backend);
  else
    b->write=lowB_write(b->backend);
  b->handles=(int *)__ogg_malloc(sizeof(int)*(1<<(b->bits)),OM_MACHINE);

  for(i=1;i<(1<<(b->bits));i++)
    b->handles[i]=(*(lowB_create(b->backend)))
      (i+1,__ogg_fixedARG(m,2),__ogg_fixedARG(m,3),__ogg_fixedARG(m,4));

  /* No action, but we do have a softreset */
  m->softreset=Bdpl_softreset;
}


/******** Bdpc ***************************************************/

/* m bits X n bits */

typedef struct Bdpc{
  int *handles;
  int words;
  int vals;

  module *backend;
  void (*write)(int,int,int,int);
  int (*read)(int,int);
}Bdpc;

static void Bdpc_write(long val){
  int word;
  Bdpc *b=(Bdpc *)__ogg_Bstate;
  word=__ogg_Bmax;
  (*(b->write))(b->handles[word],val,1,1);
}

static long Bdpc_read(void){
  int word,val;
  Bdpc *b=(Bdpc *)__ogg_Bstate;
  word=__ogg_Bmax;
  val=(*(b->read))(b->handles[word],1);
  return(val);
}

static void Bdpc_softreset(module *m){
  int i;
  Bdpc *b=(Bdpc *)m->state;

  for(i=1;i<b->words;i++)
    (*(lowB_softreset(b->backend)))(b->handles[i]);
}

static void Bdpc_undo(void){
  int i;
  Bdpc *b=(Bdpc *)__ogg_Bstate;

  for(i=0;i<b->words;i++)
    (*(lowB_undo(b->backend)))(b->handles[i]);
}

static void Bdpc_commit(void){
  int i;
  Bdpc *b=(Bdpc *)__ogg_Bstate;

  for(i=0;i<b->words;i++)
    (*(lowB_commit(b->backend)))(b->handles[i]);
}

/* Bdpc     keybitsA valbitsA backendM initA maxA rwA; */

void __oggc_Bdpc(module *m){
  Bdpc *b=(Bdpc *)(m->state=__ogg_calloc(1,sizeof(Bdpl),OM_MACHINE));
  int i;
  b->words=1<<__ogg_fixedARG(m,0);
  b->vals=1<<__ogg_fixedARG(m,1);
  b->backend=__ogg_gethandleM(__ogg_fixedARG(m,2));
  if(__ogg_fixedARG(m,5))
    b->read=lowB_read(b->backend);
  else
    b->write=lowB_write(b->backend);
  b->handles=(int *)__ogg_malloc(sizeof(int)*b->words,OM_MACHINE);

  for(i=0;i<b->words;i++)
    b->handles[i]=(*(lowB_create(b->backend)))
      (b->vals,__ogg_fixedARG(m,3),__ogg_fixedARG(m,4),__ogg_fixedARG(m,5));

  /* No action, but we do have a softreset */
  m->softreset=Bdpc_softreset;
}


/*****************************************************************/
/* TOPLEVEL function registry                                    */
/*****************************************************************/

/* We need a call registry; the top level B routines are all meant to be
   imported into other modules.  We need an easy way to tell if we're
   importing a module that actually fits, as well as returning the
   pointers to the various necessary calls at initialization time in
   order to improve performance runtime. */

/* These routines re for top level (externally available routines) */

void (*__ogg_B_write_p(module *m))(long){
  /* Gets the function pointer for a write call from a v/m pair
     routine */
  if(!strcmp(m->name,"B"))/* Straight write 'B' module */
    return(B_write);
  if(!strcmp(m->name,"Bdpl"))
    return(Bdpl_write);
  if(!strcmp(m->name,"Bdpc"))
    return(Bdpc_write);

  __ogg_Cerror(104,0); /* Gee, must not have been a compatible function */
  return(NULL); 
}

long (*__ogg_B_read_p(module *m))(void){
  /* Gets the function pointer for a read call from a v/m pair
     routine */
  if(!strcmp(m->name,"B"))/* Straight write 'B' module */
    return(B_read);
  if(!strcmp(m->name,"Bdpl"))
    return(Bdpl_read);
  if(!strcmp(m->name,"Bdpc"))
    return(Bdpc_read);

  __ogg_Cerror(104,0); /* Gee, must not have been a compatible function */
  return(NULL); 
}

void (*__ogg_B_commit_p(module *m))(void){
  /* Gets the function pointer to commit dictionary changes */

  if(!strcmp(m->name,"B"))
    return(NULL);
  if(!strcmp(m->name,"Bdpl"))
    return(Bdpl_commit);
  if(!strcmp(m->name,"Bdpc"))
    return(Bdpc_commit);

  __ogg_Cerror(104,0); /* Gee, must not have been a compatible function */
  return(NULL); 
}

void (*__ogg_B_undo_p(module *m))(void){
  /* Gets the function pointer to undo dictionary changes since last
     commit */

  if(!strcmp(m->name,"B"))
    return(NULL);
  if(!strcmp(m->name,"Bdpl"))
    return(Bdpl_undo);
  if(!strcmp(m->name,"Bdpc"))
    return(Bdpc_undo);

  __ogg_Cerror(104,0); /* Gee, must not have been a compatible function */
  return(NULL); 
}

static void __ogga_Bxc(module *m){
  module *backend=__ogg_gethandleM(__ogg_fixedARG(m,0));
  __ogg_Bstate=backend->state;
  (*(__ogg_B_commit_p(backend)))();
}

void __oggc_Bxc(module *m){
  m->action=__ogga_Bxc;
}

static void __ogga_Bxu(module *m){
  module *backend=__ogg_gethandleM(__ogg_fixedARG(m,0));
  __ogg_Bstate=backend->state;
  (*(__ogg_B_undo_p(backend)))();
}

void __oggc_Bxu(module *m){
  m->action=__ogga_Bxu;
}
