/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: F.c
  function: Basic FFT and frequency domain transform toplevels
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Jul 31 1996

 ********************************************************************/

#include <math.h>
#include "ogg.h"
#include "memory.h"
#include "parser.h"
#include "virtual.h"
#include "fft.h"
#include "ilog.h"
#include "F.h"

extern int OM_MACHINE;
extern int __ogg_precision;

/* This caches lookup information; for that reason, use a different
   transform module for each expected transform size */

typedef struct F{
  int ip;
  int op;

  long cachedsize;
  long *itrigcache;
  float *trigcache;
  int *splitcache;
  void (*itransform)(int,long *,long *,int *);
  void (*transform)(int,float *,float *,int *);
  void (*iinitcache)(int,long *,int *);
  void (*initcache)(int,float *,int *);

  long timerange;
  long scale;
  long scaled;
  int sizeoverride;
} F;

/* Fxx    pipeinP pipeoutP */

static void __ogga_F(module *m){
  F *f=(F *)m->state;
  void *b=__ogg_copyfromP(f->ip);
  long size=__ogg_sizeMB(b),i;
  int channels=__ogg_chanMB(b),ch;
  int scalen=f->sizeoverride?__ogg_fixedARG(m,2):1;
  int scaled=f->sizeoverride?__ogg_fixedARG(m,3):size;
  float *floatptr;

  if(scaled==0)scaled=1;
  scaled=f->scaled?scaled:1;

  if(size==0){
    __ogg_resetP(f->op);
    return;
  }
  if(f->cachedsize!=size){
    __ogg_free(f->trigcache);
    __ogg_free(f->itrigcache);
    __ogg_free(f->splitcache);
    f->cachedsize=size;

    /* Proper size for the largest cache requirements */

    f->trigcache=
      (float *)__ogg_calloc(3*size,sizeof(float),OM_MACHINE);

    if(f->iinitcache)
      f->itrigcache=(long *)__ogg_calloc(3*size,sizeof(long),OM_MACHINE);

    f->splitcache=(int *)__ogg_calloc(30,sizeof(int),OM_MACHINE);
    (*(f->initcache))(size,f->trigcache,f->splitcache);
    if(f->iinitcache)
      (*(f->iinitcache))(size,f->itrigcache,f->splitcache);
  }
  
  if(f->itransform  
     && __ogg_precision<0 ){

    /* We're going to do a fast integer version of the transform.
       a small fractional component is useful... */
    int fracbits=4;
    int fracmult=1<<fracbits;
    
    for(ch=0;ch<channels;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      for(i=0;i<size;i++)ptr[i]*=fracmult;

      (*(f->itransform))(size,ptr,f->itrigcache,f->splitcache);

      if(f->scale){
	long temp=f->scale*scaled,adj;
	int shift=strict_power_of_2(temp);
	if(shift>=0){
	  int temp=strict_power_of_2(scalen);
	  shift+=fracbits;
	  if(temp>=0){
	    shift-=temp;
	    if(shift)
	      adj=1<<(shift-1);
	    else
	      adj=0;
	    for(i=0;i<size;i++)
	      if(ptr[i]<0)
		ptr[i]=(ptr[i]+adj)>>shift;
	      else
		ptr[i]=-((-ptr[i]+adj)>>shift);

	  }else{
	    if(shift)
	      adj=1<<(shift-1);
	    else
	      adj=0;
	    
	    for(i=0;i<size;i++)
	      if(ptr[i]<0)
		ptr[i]=(ptr[i]*scalen+adj)>>shift;
	      else
		ptr[i]=-((-ptr[i]*scalen+adj)>>shift);

	  }
	}else
	  for(i=0;i<size;i++)
	    ptr[i]=ptr[i]*scalen/((f->scale*scaled)<<fracbits);
      }else{
	int adj=1<<(fracbits-1);
	for(i=0;i<size;i++)
	  if(ptr[i]<0)
	    ptr[i]=(ptr[i]+adj)>>fracbits;
	  else
	    ptr[i]=-((-ptr[i]+adj)>>fracbits);
      }
    }
  }else{
    floatptr=__ogg_calloc(size,sizeof(float),OM_TEMP);
    for(ch=0;ch<channels;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      for(i=0;i<size;i++)floatptr[i]=ptr[i];
      (*(f->transform))(size,floatptr,f->trigcache,f->splitcache);
      if(f->scale){
	float temp=(float)scalen/(f->scale*scaled);
	for(i=0;i<size;i++)ptr[i]=rint(floatptr[i]*temp);
      }else
	for(i=0;i<size;i++)ptr[i]=rint(floatptr[i]);
    }
    __ogg_free(floatptr);
  }

  __ogg_sendtoP(f->op,b);

}

void __oggc_Ff(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->initcache=__ogg_fdrffti;
  f->transform=__ogg_fdrfftf;
  f->scale=0;
  m->action=__ogga_F;
}

void __oggc_Ffi(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->initcache=__ogg_fdrffti;
  f->transform=__ogg_fdrfftb;
  f->scale=1;
  f->scaled=1;
  m->action=__ogga_F;
}

void __oggc_Fcq(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->initcache=__ogg_fdcosqi;
  f->transform=__ogg_fdcosqf;
  f->scale=2;
  f->scaled=0;
  m->action=__ogga_F;
}

void __oggc_Fcqi(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->initcache=__ogg_fdcosqi;
  f->transform=__ogg_fdcosqb;
  f->scale=2;
  f->scaled=1;
  m->action=__ogga_F;
}

void __oggc_Ffs(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->sizeoverride=1;
  f->initcache=__ogg_fdrffti;
  f->transform=__ogg_fdrfftf;
  f->scale=1;
  f->scaled=1;
  m->action=__ogga_F;
}

void __oggc_Ffis(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->sizeoverride=1;
  f->initcache=__ogg_fdrffti;
  f->transform=__ogg_fdrfftb;
  f->scale=1;
  f->scaled=1;
  m->action=__ogga_F;
}

void __oggc_Fcqs(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->sizeoverride=1;
  f->initcache=__ogg_fdcosqi;
  f->transform=__ogg_fdcosqf;
  f->scale=2;
  f->scaled=1;
  m->action=__ogga_F;
}

void __oggc_Fcqis(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->sizeoverride=1;
  f->initcache=__ogg_fdcosqi;
  f->transform=__ogg_fdcosqb;
  f->scale=2;
  f->scaled=1;
  m->action=__ogga_F;
}

void __oggc_Ff32(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->initcache=__ogg_fdrffti;
  f->transform=__ogg_fdrfftf;
  f->iinitcache=__ogg_idrffti;
  f->itransform=__ogg_idrfftf;
  f->timerange=8;
  f->scale=0;
  m->action=__ogga_F;
}

void __oggc_Ffi32(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->initcache=__ogg_fdrffti;
  f->transform=__ogg_fdrfftb;
  f->iinitcache=__ogg_idrffti;
  f->itransform=__ogg_idrfftb;
  f->timerange=8;
  f->scale=1;
  f->scaled=1;
  m->action=__ogga_F;
}

void __oggc_Fcq32(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->initcache=__ogg_fdcosqi;
  f->transform=__ogg_fdcosqf;
  f->iinitcache=__ogg_idrffti;
  f->itransform=__ogg_idrfftf;
  f->timerange=8;
  f->scale=2;
  f->scaled=0;
  m->action=__ogga_F;
}

void __oggc_Fcqi32(module *m){
  F *f=(F *)(m->state=__ogg_calloc(1,sizeof(F),OM_MACHINE));
  f->ip=__ogg_fixedARG(m,0);
  f->op=__ogg_fixedARG(m,1);
  f->initcache=__ogg_fdcosqi;
  f->transform=__ogg_fdcosqb;
  f->iinitcache=__ogg_idrffti;
  f->itransform=__ogg_idrfftb;
  f->timerange=8;
  f->scale=2;
  f->scaled=1;
  m->action=__ogga_F;
}

static void __ogga_Fn(module *m){
  int p=__ogg_fixedARG(m,0);
  void *b=__ogg_copyfromP(p);
  long size=__ogg_sizeMB(b),i;
  int channels=__ogg_chanMB(b),ch;
  float temp=size?1/size:1;

  for(ch=0;ch<channels;ch++){
    long *ptr=__ogg_ptrMB(b,ch);
    for(i=0;i<size;i++)
      ptr[i]*=temp*(i+1);
  }

  __ogg_sendtoP(p,b);
}

void __oggc_Fn(module *m){
  m->action=__ogga_Fn;
}

/* Quantizers and mappers (perceptual here too) */

static long *foo_weights;

#if defined(__GNUC__) || defined( WIN32 ) || defined( MACVER )
static int foo_compare(const void *a,const void *b){
#else
static int foo_compare(void *a,void *b){
#endif
  return(labs(foo_weights[*((long *)b)])-labs(foo_weights[*((long *)a)]));
}

/* This one isn't perceptual */
static void __ogga_Fm(module *m){
  int p=__ogg_fixedARG(m,0);
  int freq=__ogg_fixedARG(m,1);
  int map=__ogg_fixedARG(m,2);
  int from=__ogg_fixedARG(m,3);
  int to=__ogg_fixedARG(m,4);
  void *in=__ogg_readfromP(p);
  long size=__ogg_sizeMB(in),i,j;
  int chan=__ogg_chanMB(in),ch;
  long mapto=0;

  __ogg_resetP(freq);
  __ogg_resetP(map);

  if(in){
    void *fout=NULL,*m=__ogg_allocMB(chan,size);
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(in,ch);
      long *mapptr=__ogg_ptrMB(m,ch);
      long *out;
      long *index=__ogg_malloc(sizeof(long)*from,OM_TEMP);

      mapto=0;
      foo_weights=ptr;

      for(i=0;i<size;i+=from){
	if(i+from>size){
	  /* Boundary condition */
	  for(j=i;j<size;j++){
	    mapto++;
	    mapptr[j]=-1;
	  }
	}else{
	  for(j=0;j<from;j++)index[j]=i+j;
	  qsort(index,from,sizeof(long),foo_compare);
	  for(j=0;j<to;j++){
	    mapto++;
	    mapptr[index[j]]=-1;
	  }
	}
      }

      if(!fout)fout=__ogg_allocMB(chan,mapto);
      out=__ogg_ptrMB(fout,ch);

      mapto=0;
      for(i=0;i<size;i++)if(mapptr[i])out[mapto++]=ptr[i];
    __ogg_free(index);
    }

    __ogg_sendtoP(p,in);
    __ogg_sendtoP(freq,fout);
    __ogg_sendtoP(map,m);

  }
}

static void __ogga_Fmc(module *m){
  int p=__ogg_fixedARG(m,0);
  int freq=__ogg_fixedARG(m,1);
  int map=__ogg_fixedARG(m,2);
  int from=__ogg_fixedARG(m,3);
  int to=__ogg_fixedARG(m,4);
  void *in=__ogg_readfromP(p);
  long size=__ogg_sizeMB(in),i,j;
  int chan=__ogg_chanMB(in),ch;
  long mapto=0;

  __ogg_resetP(freq);
  __ogg_resetP(map);

  if(in){
    void *fout=NULL,*m=__ogg_allocMB(chan,size/2);
    foo_weights=__ogg_malloc(size/2*sizeof(long),OM_TEMP);

    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(in,ch);
      long *mapptr=__ogg_ptrMB(m,ch);
      long *out;
      long *index; 

      index=__ogg_malloc(sizeof(long)*from,OM_TEMP);

      for(i=0;i+1<size;i+=2)
	foo_weights[i>>1]=hypot(ptr[i],ptr[i+1]);

      mapto=0;

      for(i=0;i<size/2;i+=from){
	if(i+from>size/2){
	  /* Boundary condition */
	  for(j=i;j<size/2;j++){
	    mapto++;
	    mapptr[j]=-1;
	  }
	}else{
	  for(j=0;j<from;j++)index[j]=i+j;
	  qsort(index,from,sizeof(long),foo_compare);
	  for(j=0;j<to;j++){
	    mapto++;
	    mapptr[index[j]]=-1;
	  }
	}
      }

      if(!fout)fout=__ogg_allocMB(chan,mapto*2);
      out=__ogg_ptrMB(fout,ch);

      mapto=0;
      for(i=0;i+1<size;i+=2)
	if(mapptr[i>>1]){
	  out[mapto++]=ptr[i];
	  out[mapto++]=ptr[i+1];
	}

      __ogg_free(index);
    }
    __ogg_free(foo_weights);
    __ogg_sendtoP(p,in);
    __ogg_sendtoP(freq,fout);
    __ogg_sendtoP(map,m);

  }
}

/* Mapping, not expansion.  Expansion belongs to the vector ops */
static void __ogga_Fmi(module *m){
  int p=__ogg_fixedARG(m,0);
  int freq=__ogg_fixedARG(m,1);
  int map=__ogg_fixedARG(m,2);
  int by=__ogg_fixedARG(m,3);

  void *in=__ogg_readfromP(p);
  void *mb=__ogg_readfromP(map);
  long size=__ogg_sizeMB(mb),i,j;
  int chan=__ogg_chanMB(mb),ch;
  long smallsize=__ogg_sizeMB(in);

  if(by<1)by=1;

  __ogg_resetP(freq);

  if(in && mb){
    void *fout=__ogg_allocMB(chan,size*by);
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(in,ch);
      long *map=__ogg_ptrMB(mb,ch);
      long *out=__ogg_ptrMB(fout,ch);

      long mapptr=0;

      switch(by){
      case 1:
	for(i=0;i<size && mapptr<smallsize;i++)
	  if(map[i]<0)out[i]=ptr[mapptr++];
	break;
      case 2:
	for(i=0;i<size && mapptr+1<smallsize;i++)
	  if(map[i]<0){
	    out[i<<1]=ptr[mapptr++];
	    out[(i<<1)+1]=ptr[mapptr++];
	  }
	break;
      default:
	for(i=0;i<size && mapptr<smallsize;i++)
	  for(j=0;j<by && mapptr<smallsize;j++)
	    out[i*by+j]=ptr[mapptr++]; /* Optimize later; not used yet */
      }
    }
    
    __ogg_sendtoP(p,in);
    __ogg_sendtoP(freq,fout);
    __ogg_sendtoP(map,mb);
  }

}

void __oggc_Fm(module *m){
  m->action=__ogga_Fm;
}

void __oggc_Fmc(module *m){
  m->action=__ogga_Fmc;
}

void __oggc_Fmi(module *m){
  m->action=__ogga_Fmi;
}






