/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: S.c
  function: Miscellaneous support modules
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Aug 09 1996

 ********************************************************************/

#include "ogg.h"
#include "memory.h"
#include "parser.h"
#include "virtual.h"
#include "bitwise.h"
#include "ilog.h"
#include "lpc.h"
#include "S.h"
#include <math.h>

/********* S{oi}f ******************************************************/

/* The simplest input buffer -> pipe filter.  It just grabs a fixed
   size frame (evaluated at creation time); if there's not enough in
   the buffer, it grabs what it can.  It stores the grab size in a
   var, which should be used later for bitstream terminate detect. */

extern int OM_MACHINE;

typedef struct Sf{
  int inpipe;
  int outpipe;
  int bpipe;
  int outvar;
  int fetch;
} Sf;

static void __ogga_Sof(module *m){
  Sf *s=(Sf *)(m->state);
  long available=__ogg_testP(s->inpipe);

  __ogg_Bset(s->bpipe);

  if(available<s->fetch){
    __ogg_Bout(0,1);
    __ogg_Bout(available,ilog(s->fetch));
    __ogg_sendtoP(s->outpipe,__ogg_readfromP(s->inpipe));
    __ogg_writetoV(s->outvar,available);
  }else{
    __ogg_Bout(1,1);
    if(available==s->fetch)
      __ogg_sendtoP(s->outpipe,__ogg_readfromP(s->inpipe));
    else
      __ogg_sendtoP(s->outpipe,__ogg_readnfromP(s->inpipe,s->fetch));
    __ogg_writetoV(s->outvar,s->fetch);
  }
}

/* soft reset is a nop */

void __oggc_Sof(module *m){
  Sf *s=(Sf *)(m->state=__ogg_calloc(1,sizeof(Sf),OM_MACHINE));

  s->inpipe=__ogg_fixedARG(m,0);
  s->outpipe=__ogg_fixedARG(m,1);
  s->bpipe=__ogg_fixedARG(m,2);
  s->outvar=__ogg_fixedARG(m,4);
  s->fetch=__ogg_fixedARG(m,3);

  m->action=__ogga_Sof;
}

static void __ogga_Sif(module *m){
  Sf *s=(Sf *)(m->state);

  __ogg_Bset(s->bpipe);
  if(__ogg_Bread(1))
    __ogg_writetoV(s->outvar,s->fetch);
  else
    __ogg_writetoV(s->outvar,__ogg_Bread(ilog(s->fetch)));
}

/* soft reset is a nop */

void __oggc_Sif(module *m){
  Sf *s=(Sf *)(m->state=__ogg_calloc(1,sizeof(Sf),OM_MACHINE));

  s->bpipe=__ogg_fixedARG(m,0);
  s->outvar=__ogg_fixedARG(m,2);
  s->fetch=__ogg_fixedARG(m,1);

  m->action=__ogga_Sif;
}

/*** S{o|i}fr *********************************************************/
/* Sofr     inP outP outB granA maxA attackperA decayperA ordA frameV endV */

/* Rolling analysis of the input data, choosing an input length
   consisting of homogenous data */

static void __ogga_Sofr(module *m){
  int inP=__ogg_fixedARG(m,0);
  int outP=__ogg_fixedARG(m,1);
  int outB=__ogg_fixedARG(m,2);
  long granA=__ogg_fixedARG(m,3);
  long maxA=__ogg_fixedARG(m,4);
  long attackA=__ogg_fixedARG(m,5);
  long decayA=__ogg_fixedARG(m,6);
  long orderA=__ogg_fixedARG(m,7);
  int frameV=__ogg_fixedARG(m,8);
  int endV=__ogg_fixedARG(m,9);

  void *in=__ogg_readfromP(inP),*out;
  long available=__ogg_sizeMB(in),i,j;
  long chan=__ogg_chanMB(in),ch;

  long count=0,sofar=0;
  double *work1=(double *)__ogg_calloc(orderA+1,sizeof(double),OM_TEMP);
  double **work2=(double **)__ogg_malloc(chan*sizeof(double *),OM_TEMP);

  double addpower,y;
  double *lastpower=(double *)__ogg_calloc(chan,sizeof(double),OM_TEMP);

  if(granA<orderA)granA=orderA;

  while(sofar+orderA<available){
    int proc=granA;
    int start=0;
    
    /* If this is the first subframe for analysis, draw the priming
       data from the beginning of the block */
    if(count==0)start=orderA;

    /* make sure we're not screaming off past the end... */
    if(sofar+proc>available)proc=available-sofar;
    
    /* Analyze channels seperately */
    for(ch=0;ch<chan;ch++){
      /* Offset past the subframes we've already done */
      long *data=__ogg_ptrMB(in,ch)+sofar;
      
      /* Make poles, but only at the beginning */
      if(count==0){
	work2[ch]=(double *)__ogg_calloc(orderA,sizeof(double),OM_TEMP);

	autocorrelation(proc,data,orderA+1,work1);
	levinson_durbin(work1,work2[ch],orderA);
      }

      /* Get residue for comparison */
      addpower=0;
      data-=orderA; /* Take priming from the previous subframe; 
		       compensated for first frame */
      for(i=start;i<proc;i++){
	y=0;
	for(j=0;j<orderA;j++)
	  y+=data[i+j]*work2[ch][orderA-j-1];

	addpower+=fabs((data+orderA)[i]+y);
      }

      addpower/=(proc-start);

      /* Do the actual comparison now. */
      if(count>0){
	/* Have we surpassed allowable attack? decay? */
	if(lastpower[ch]*attackA/100<addpower ||
	   lastpower[ch]*decayA/100>addpower){

	  /* Breakpoint!  Deallocate temps, set up pipes and return */

	  __ogg_free(work1);
	  for(ch=0;ch<chan;ch++)
	    __ogg_free(work2[ch]);
	  __ogg_free(work2);
	  __ogg_free(lastpower);
	  
	  __ogg_appendtoP(inP,in);
	  out=__ogg_readnfromP(inP,sofar);
	  __ogg_sendtoP(outP,out);
	  __ogg_writetoV(frameV,sofar);

	  /* We know that the output frame is not the last data,
             because we just found a breakpoint within valid data,
             thus there's more to come next frame */
	  __ogg_writetoV(endV,0);
	  __ogg_Bset(outB);
	  __ogg_Bout(0,1);
	  __ogg_Bout(count-1,ilog((maxA+granA-1)/granA-1));
	  return;
	}
      }

      lastpower[ch]=addpower;
    }

    sofar+=granA;
    count++;
  }

  /* No breakpoints in submitted data.  Dealloc temps, move pipe data,
     test for end of data */

  __ogg_free(work1);
  for(ch=0;ch<chan;ch++)
    __ogg_free(work2[ch]);
  __ogg_free(work2);
  __ogg_free(lastpower);
  __ogg_sendtoP(outP,in);
  __ogg_writetoV(frameV,available);

  __ogg_Bset(outB);

  /* last frame? */

  if(available<maxA){
    __ogg_Bout(1,1);
    __ogg_Bout(available,ilog(maxA-1));
    if(available==0)__ogg_virtual_machine_halt();
    __ogg_writetoV(endV,1);
  }else{
    __ogg_Bout(0,1);
    __ogg_Bout((maxA+granA-1)/granA-1,ilog((maxA+granA-1)/granA-1));
    __ogg_writetoV(endV,0);
  }
}

void __oggc_Sofr(module *m){
  m->action=__ogga_Sofr;
}

/* Sifr     inB granA maxA frameV endV                             */

static void __ogga_Sifr(module *m){
  int inB=__ogg_fixedARG(m,0);
  long granA=__ogg_fixedARG(m,1);
  long maxA=__ogg_fixedARG(m,2);
  int frameV=__ogg_fixedARG(m,3);
  int endV=__ogg_fixedARG(m,4);

  __ogg_Bset(inB);
  if(__ogg_Bread(1)){
    long available=__ogg_Bread(ilog(maxA-1));
    __ogg_writetoV(endV,1);
    __ogg_writetoV(frameV,available);
    if(available==0)__ogg_virtual_machine_halt();
  }else{
    int count=__ogg_Bread(ilog((maxA+granA-1)/granA-1))+1;
    long available=granA*count;
    if(available>maxA)available=maxA;

    __ogg_writetoV(frameV,available);
    __ogg_writetoV(endV,0);
  }
}

void __oggc_Sifr(module *m){
  m->action=__ogga_Sifr;
}

static void __ogga_Shalt(module *m){
  if(__ogg_fixedARG(m,0)!=__ogg_fixedARG(m,1))__ogg_virtual_machine_halt();
}

/* soft reset is a nop */

void __oggc_Shalt(module *m){
  m->action=__ogga_Shalt;
}

static void __ogga_Spm(module *m){
  int inP=__ogg_fixedARG(m,0);
  int outP=__ogg_fixedARG(m,1);
  __ogg_sendtoP(outP,__ogg_readfromP(inP));
}

void __oggc_Spm(module *m){
  m->action=__ogga_Spm;
}

static void __ogga_Spc(module *m){
  int inP=__ogg_fixedARG(m,0);
  int outP=__ogg_fixedARG(m,1);
  __ogg_sendtoP(outP,__ogg_copyfromP(inP));
}

void __oggc_Spc(module *m){
  m->action=__ogga_Spc;
}

/* Move contents from one bpipe to another */
static void __ogga_Sbm(module *m){
  int inB=__ogg_fixedARG(m,0);
  int outB=__ogg_fixedARG(m,1);
  __ogg_Bmove(outB,inB);
}

void __oggc_Sbm(module *m){
  m->action=__ogga_Sbm;
}

static void __ogga_Sbc(module *m){
  int inB=__ogg_fixedARG(m,0);
  int outB=__ogg_fixedARG(m,1);
  __ogg_Bcopy(outB,inB);
}

void __oggc_Sbc(module *m){
  m->action=__ogga_Sbc;
}

static void __ogga_Sba(module *m){
  int inB=__ogg_fixedARG(m,0);
  int outB=__ogg_fixedARG(m,1);
  __ogg_Bappend(outB,inB);
}

void __oggc_Sba(module *m){
  m->action=__ogga_Sba;
}

static void __ogga_Sbt(module *m){
  int inB=__ogg_fixedARG(m,0);
  int outV=__ogg_fixedARG(m,1);
  __ogg_writetoV(outV,__ogg_Btestbits(inB));
}

void __oggc_Sbt(module *m){
  m->action=__ogga_Sbt;
}

static void __ogga_Sbr(module *m){
  int inB=__ogg_fixedARG(m,0);
  __ogg_Breset(inB);
}

void __oggc_Sbr(module *m){
  m->action=__ogga_Sbr;
}

static void __ogga_Sv(module *m){
  int in=__ogg_fixedARG(m,0);
  int outV=__ogg_fixedARG(m,1);
  __ogg_writetoV(outV,in);
}

void __oggc_Sv(module *m){
  m->action=__ogga_Sv;
}

static void __ogga_Spush(module *m){
  int inA=__ogg_fixedARG(m,0);
  int outP=__ogg_fixedARG(m,1);
  void *out=__ogg_allocMB(1,1);
  long *ptr=__ogg_ptrMB(out,0);
  ptr[0]=inA;
  __ogg_appendtoP(outP,out);
}

void __oggc_Spush(module *m){
  m->action=__ogga_Spush;
}

static void __ogga_Spull(module *m){
  int inP=__ogg_fixedARG(m,0);
  int outV=__ogg_fixedARG(m,1);
  void *in=__ogg_readnfromP(inP,1);
  long *ptr=__ogg_ptrMB(in,0);
  if(ptr)
    __ogg_writetoV(outV,ptr[0]);
  __ogg_freeMB(in);
}

void __oggc_Spull(module *m){
  m->action=__ogga_Spull;
}

