/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: bitwise.c
  function: bitpacking routines
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Aug 03 1996

 ********************************************************************/

/* Different from what we had pre 0.98; this does no file I/0.  We
   merely pack values into a 8-bit byte aligned buffer.  Why 8 bits?
   Because that's the largest common word size in modern computer
   systems that is not affected by endianness issues. */

/* Also different, the bitpacking is implemented as a VOM type: 'B'
   for 'bitpipe', like the Ogg_pipes. */

/* This is the fastest fully portable bitpacker I could write... to go
   any farther, you'd probably need to take advantage of:

   a) assembly
   b) 64 bit math (maybe)
   c) larger data types, taking into account machine specific
      endianness and alignment issues (maybe)
   d) hardware
   e) flattening out functional calls (make these functions inline)
   
   It assumes that most of our writes will be around 4-10 bits long, and
   is tuned for such.  This is about right for Ogg.*/

/* Attention!  We're 'LSb' endian; if we write a word but read
   individual bits, then we'll read the lsb first */

#include "ogg.h"
#include "memory.h"
#include "memtypes.h"
#include "bitwise.h"

#ifdef SELFTESTS
#  include <stdio.h> /* For reporting errors */
#  include "ilog.h"
#  define BUFFER_INCREMENT 5
   int ogg_errno;
   jmp_buf __ogg_jmp_env;
   int __ogg_jmp_env_set=0;
#else
#  define BUFFER_INCREMENT 1024
   extern int ogg_errno;
   extern jmp_buf __ogg_jmp_env;
   extern int __ogg_jmp_env_set;
#endif

bpipe **__ogg_bpipelist=NULL;
int __ogg_bpipes=0;
static bpipe *Bdirect;

#ifndef SELFSUFF
extern int OM_BPIPE;
#endif

void __ogg_cleanupB(void){
  __ogg_bpipelist=NULL;
  __ogg_bpipes=0;
  __ogg_ripfree(OM_BPIPE);
}

static void longret(void){
  if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
  else exit(1);
}

static bpipe *__ogg_balloc(long block){
  return((bpipe *)__ogg_calloc(1,sizeof(bpipe),block));
}

/* Register a bpipe, return -1 on error */
void __ogg_Bregister(int pipe){
  int i;

  if(pipe<0){
    ogg_errno=5;
    longret();
  }

  if(pipe>=__ogg_bpipes){
    __ogg_bpipelist=(bpipe **)__ogg_realloc(__ogg_bpipelist,
					    sizeof(bpipe *)*(pipe+1),
					    OM_BPIPE);
    for(i=__ogg_bpipes;i<=pipe;i++)__ogg_bpipelist[i]=NULL;
    __ogg_bpipes=pipe+1;
  }
  if(__ogg_bpipelist[pipe]==NULL)
    __ogg_bpipelist[pipe]=__ogg_balloc(OM_BPIPE);
}

/* Alloc a bpipe in the OM_TEMP group and set Bdirect to use it */
void __ogg_Btempalloc(void){
  bpipe *ret=__ogg_balloc(OM_TEMP);
  Bdirect=ret;
}

STIN bpipe *bget(int pipe){
  if(pipe==-1)
    return(Bdirect);
  else
    return(__ogg_bpipelist[pipe]);
}

void __ogg_Breset(int pipe){
  bpipe *b=bget(pipe);
  __ogg_free(b->buffer);
  b->buffer=NULL;
  if(b->bytes>b->biggest)b->biggest=b->bytes;
  b->bytes=0;
  b->endbyte=0;
  b->endbit=0;
}

/* This *copies* the input into OM_BPIPE */
int __ogg_Bfill(int pipe,char *in,long bytes){
  bpipe *b=bget(pipe);
  __ogg_free(b->buffer);
  b->buffer=NULL;
  if(b->bytes>b->biggest)b->biggest=b->bytes;
  b->bytes=0;
  b->endbyte=0;
  b->endbit=0;
  b->buffer=(unsigned char *)__ogg_malloc(bytes+5,OM_BPIPE);
  if(bytes)memcpy(b->buffer,in,bytes);
  memset(b->buffer+bytes,0,5);
  b->bytes=bytes;
  return(0);
}

void __ogg_Bappend(int dest, int source){
  bpipe *d=bget(dest);
  bpipe *s=bget(source);

  if(d->bytes<d->endbyte+s->endbyte+2)
    if(dest==-1)
      d->buffer=__ogg_realloc(d->buffer,d->endbyte+s->endbyte+2,OM_TEMP);
    else
      d->buffer=__ogg_realloc(d->buffer,d->endbyte+s->endbyte+2,OM_BPIPE);

  d->bytes=d->endbyte+s->endbyte+2;

  if(d->endbit==0){
    memcpy(d->buffer+d->endbyte,s->buffer,s->endbyte);
    if(s->endbit)
      d->buffer[d->endbyte+s->endbyte]=s->buffer[s->endbyte];
    d->endbyte+=s->endbyte;
    d->endbit=s->endbit;
  }else{
    int i;
    int bytes=s->endbyte;
    if(s->endbit)bytes++;
    for(i=0;i<bytes;i++){
      d->buffer[d->endbyte+i]|=((unsigned char)(s->buffer[i]))<<d->endbit;
      d->buffer[d->endbyte+i+1]=((unsigned char)(s->buffer[i]))>>(8-d->endbit);
    }
    d->endbyte+=s->endbyte;
    d->endbit+=s->endbit;
    if(d->endbit>7){
      d->endbit-=8;
      d->endbyte++;
    }
  }
  __ogg_Breset(source);
}

void __ogg_Bcopy(int dest, int source){
  bpipe *d=bget(dest);
  bpipe *s=bget(source);

  __ogg_Breset(dest);

  d->bytes=s->bytes;
  d->endbyte=s->endbyte;
  d->endbit=s->endbit;

  if(dest==-1)
    d->buffer=__ogg_malloc(s->bytes,OM_TEMP);
  else
    d->buffer=__ogg_malloc(s->bytes,OM_BPIPE);

  if(d->bytes)memcpy(d->buffer,s->buffer,d->bytes);
}

void __ogg_Bmove(int dest, int source){
  bpipe *d=bget(dest);
  bpipe *s=bget(source);

  __ogg_Breset(dest);

  d->bytes=s->bytes;
  d->endbyte=s->endbyte;
  d->endbit=s->endbit;

  /* we might be moving to or from a temp pipe */
  if(dest==-1)
    __ogg_ripmove(d->buffer=s->buffer,OM_TEMP); 
  else
    if(source==-1)
      __ogg_ripmove(d->buffer=s->buffer,OM_BPIPE); 
    else
      d->buffer=s->buffer;

  s->buffer=NULL;
  __ogg_Breset(source);
}

/* Hand back the buffer byte justified and reset the pipe */
/* Move the buffer into OM_TEMP */

void __ogg_Bwflush(int bpipenum, char **buffer, long *bytes){
  bpipe *b=bget(bpipenum);
  *buffer=(char *)(b->buffer);
  if(b->endbit)
    *bytes=b->endbyte+1;
  else
    *bytes=b->endbyte;
  b->buffer=NULL;
  __ogg_ripmove(*buffer,OM_TEMP);
  __ogg_Breset(bpipenum);
}

/* How many bytes+partial bytes did we read from the input buffer? */
long __ogg_Brflush(int bpipenum){
  int ret;
  bpipe *b=bget(bpipenum);
  if(b->endbit)
    ret=b->endbyte+1;
  else
    ret=b->endbyte;
  __ogg_Breset(bpipenum);
  return(ret);
}

/* Return what's left unused in OM_TEMP; assume byte framing alignment
   and return only wholly unread bytes. */
long __ogg_Brremaining(int bpipenum,char **rembuffer){
  int ret;
  bpipe *b=bget(bpipenum);
  char *buffer=(char *)(b->buffer);
  if(b->endbit)
    ret=b->endbyte+1;
  else
    ret=b->endbyte;
 
  ret=b->bytes-ret;
  if(ret<=0)
    *rembuffer=NULL;
  else{
    *rembuffer=(char *)__ogg_malloc(ret,OM_TEMP);
    memcpy(*rembuffer,buffer+(b->bytes-ret),ret);
  }
  __ogg_Breset(bpipenum);
  return(ret);
}

void __ogg_Bset(int pipe){
  Bdirect=__ogg_bpipelist[pipe];
}

static unsigned long mask[]=
{0x00000000,0x00000001,0x00000003,0x00000007,0x0000000f,
 0x0000001f,0x0000003f,0x0000007f,0x000000ff,0x000001ff,
 0x000003ff,0x000007ff,0x00000fff,0x00001fff,0x00003fff,
 0x00007fff,0x0000ffff,0x0001ffff,0x0003ffff,0x0007ffff,
 0x000fffff,0x001fffff,0x003fffff,0x007fffff,0x00ffffff,
 0x01ffffff,0x03ffffff,0x07ffffff,0x0fffffff,0x1fffffff,
 0x3fffffff,0x7fffffff,0xffffffff };

/* Takes only up to 32 bits */
void __ogg_Bout(unsigned long value,int bits){
  unsigned char *ptr;
  if(Bdirect->endbyte+5>=Bdirect->bytes){ /* We are handling 32 bit input */
    if(Bdirect->bytes<Bdirect->biggest){
      Bdirect->buffer=(unsigned char *)__ogg_realloc(Bdirect->buffer,
						     Bdirect->biggest,
						     OM_BPIPE);
      Bdirect->bytes=Bdirect->biggest;
    }else{
      Bdirect->buffer=(unsigned char *)__ogg_realloc(Bdirect->buffer,
					      Bdirect->bytes+BUFFER_INCREMENT,
						     OM_BPIPE);
      Bdirect->bytes+=BUFFER_INCREMENT;
    }
    if(Bdirect->endbit==0)Bdirect->buffer[Bdirect->endbyte]=0;
  }

  value&=mask[bits]; /* Necessary for truncation cases */
  ptr=Bdirect->buffer+Bdirect->endbyte;

  bits+=Bdirect->endbit;
  if(bits>15){
    *(ptr++)|=(unsigned char)(value<<Bdirect->endbit);
    value>>=(8-Bdirect->endbit);
    *(ptr++)=(unsigned char)value;
    *(ptr++)=(unsigned char)(value>>8);
    *(ptr++)=(unsigned char)(value>>16);
    *ptr=(unsigned char)(value>>24);
  }else{  
    *(ptr++)|=(unsigned char)(value<<Bdirect->endbit);
    value<<=Bdirect->endbit;
    *ptr=(unsigned char)(value>>8);
  }
  Bdirect->endbyte+=bits>>3;
  Bdirect->endbit=bits&7;
}

void __ogg_BoutSl(char *s){
  int leng,count;
  
  leng=strlen(s);
  for(count=0;count<leng;count++)
    __ogg_Bout(s[count],8);
}

void __ogg_BoutS(char *s){
  __ogg_BoutSl(s);
  __ogg_Bout(0,8);
}

/* Read in bits without advancing the bitptr */
/* We have alloced enough not to fall off the end with 32+7 bit readahead */
unsigned long __ogg_Blook(int bits){
  unsigned long bmask=mask[bits],ret;
  unsigned char *ptr=Bdirect->buffer+Bdirect->endbyte;
  
  bits+=Bdirect->endbit;
  if(bits>16){
    ret=*(ptr+4);
    ret<<=8;
    ret|=*(ptr+3);
    ret<<=8;
    ret|=*(ptr+2);
    ret<<=8;
    ret|=*(ptr+1);
    ret<<=(8-Bdirect->endbit);
    ret|=(*ptr)>>Bdirect->endbit;
  }else{
    ret=*(ptr+1);
    ret<<=8;
    ret|=*ptr;
    ret>>=Bdirect->endbit;
  }
  return(bmask&ret);
}

void __ogg_Badv(int bits){
  bits+=Bdirect->endbit;
  Bdirect->endbit=bits&7;
  if((Bdirect->endbyte+=(bits>>3))<Bdirect->bytes)return;
  if(Bdirect->endbit+Bdirect->endbyte>Bdirect->bytes){
    Bdirect->endbyte=Bdirect->bytes;
    Bdirect->endbit=0;
    ogg_errno=2;
    if(__ogg_jmp_env_set){
      longjmp(__ogg_jmp_env,1);
    }else
      exit(1);
  }
  return;
}

long __ogg_Bread(int bits){
  long ret=(long)__ogg_Blook(bits);
  __ogg_Badv(bits);
  return(ret);
}

/* The string memory returned is OM_TEMP */
char *__ogg_BinS(void){
  /* This is annoying to do right :) We don't want to waste space or
     realloc for every read char... */
  char buffer[80];
  int ichar,slen=0,blen=0;
  char *strin=NULL;
  
  while(!isspace(ichar=__ogg_Bread(8))){
    if(ichar<0)return(NULL);
    buffer[blen++]=ichar;
    if(blen==80){
      strin=__ogg_realloc(strin,slen+80,OM_TEMP);
      memcpy(strin+slen,buffer,80);
      slen+=80;
      blen=0;
    }
  }
  buffer[blen++]='\0';
  strin=__ogg_realloc(strin,slen+blen,OM_TEMP);
  memcpy(strin+slen,buffer,blen);
  return(strin);
}

char *__ogg_BinSS(void){
  char buffer[80];
  int ichar,slen=0,blen=0;
  char *strin=NULL;
  
  while((ichar=__ogg_Bread(8))>0){
    buffer[blen++]=ichar;
    if(blen==80){
      strin=__ogg_realloc(strin,slen+80,OM_TEMP);
      memcpy(strin+slen,buffer,80);
      slen+=80;
      blen=0;
    }
  }
  buffer[blen++]='\0';
  strin=__ogg_realloc(strin,slen+blen,OM_TEMP);
  memcpy(strin+slen,buffer,blen);
  return(strin);
}

long __ogg_Btest(int pipe){
  bpipe *b=bget(pipe);
  if(b->endbit)
    return(b->endbyte+1);
  else
    return(b->endbyte);
}

long __ogg_Btestbits(int pipe){
  bpipe *b=bget(pipe);
  return((b->endbyte<<3)+b->endbit);
}

/* Self test of the bitpack routines; everything else is based on
   them, so they damned well better be solid. */
/* Thoroughly tests the unsigned char shifts in Blook */
#ifdef SELFTESTS

void report(char *in){
  fprintf(stderr,in);
  exit(1);
}

void cliptest(unsigned long *b,int vals,int *comp,int compsize){
  long bytes,i;
  unsigned char *buffer;

  for(i=0;i<vals;i++)
    __ogg_Bout(b[i],ilog(b[i]));
  __ogg_Bwflush(0,(char **)(&buffer),&bytes);
  if(bytes!=compsize)report("wrong wrong number of bytes!\n");
  for(i=0;i<bytes;i++)if(buffer[i]!=comp[i])report("wrote incorrect value!\n");
  __ogg_Bfill(0,(char *)buffer,bytes);
  for(i=0;i<vals;i++)
    if(__ogg_Bread(ilog(b[i]))!=b[i])report("read incorrect value!\n");
  if(__ogg_Brflush(0)!=bytes)report("leftover bytes after read!\n");
  __ogg_free(buffer);
}

void uncliptest(unsigned long *b,int vals,int bits,int *comp,int compsize){
  long bytes,i;
  unsigned char *buffer;

  for(i=0;i<vals;i++)
    __ogg_Bout(b[i],bits);
  __ogg_Bwflush(0,(char **)(&buffer),&bytes);
  if(bytes!=compsize)report("wrong wrong number of bytes!\n");
  for(i=0;i<bytes;i++)if(buffer[i]!=comp[i])report("wrote incorrect value!\n");
  __ogg_Bfill(0,(char *)buffer,bytes);
  for(i=0;i<vals;i++)
    if(__ogg_Bread(bits)!=(mask[bits]&b[i]))report("read incorrect value!\n");
  if(__ogg_Brflush(0)!=bytes)report("leftover bytes after read!\n");
  __ogg_free(buffer);
}

int main(void){
  unsigned char *buffer;
  long bytes,i;
  static unsigned long testbuffer1[]=
    {18,12,103948,4325,543,76,432,52,3,65,4,56,32,42,34,21,1,23,32,546,456,7,
       567,56,8,8,55,3,52,342,341,4,265,7,67,86,2199,21,7,1,5,1,4};
  int test1size=43;

  static unsigned long testbuffer2[]=
    {216531625L,1237861823,56732452,131,3212421,12325343,34547562,12313212,
       1233432,534,5,346435231,14436467,7869299,76326614,167548585,
       85525151,0,12321,1,349528352};
  int test2size=21;

  static unsigned long large[]=
    {2136531625L,2137861823,56732452,131,3212421,12325343,34547562,12313212,
       1233432,534,5,2146435231,14436467,7869299,76326614,167548585,
       85525151,0,12321,1,2146528352};

  static unsigned long testbuffer3[]=
    {1,0,14,0,1,0,12,0,1,0,0,0,1,1,0,1,0,1,0,1,0,1,0,1,0,1,0,0,1,1,1,1,1,0,0,1,
       0,1,30,1,1,1,0,0,1,0,0,0,12,0,11,0,1,0,0,1};
  int test3size=56;

  int onesize=33;
  static int one[]={146,25,44,151,195,15,153,176,233,131,196,65,85,172,47,40,
		    34,242,223,136,35,222,211,86,171,50,225,135,214,75,172,
		    223,4};

  int twosize=6;
  static int two[]={61,255,255,251,231,29};

  int threesize=54;
  static int three[]={169,2,232,252,91,132,156,36,89,13,123,176,144,32,254,
		      142,224,85,59,121,144,79,124,23,67,90,90,216,79,23,83,
		      58,135,196,61,55,129,183,54,101,100,170,37,127,126,10,
		      100,52,4,14,18,86,77,1};

  int foursize=38;
  static int four[]={18,6,163,252,97,194,104,131,32,1,7,82,137,42,129,11,72,
		     132,60,220,112,8,196,109,64,179,86,9,137,195,208,122,169,
		     28,2,133,0,1};

  int fivesize=45;
  static int five[]={169,2,126,139,144,172,30,4,80,72,240,59,130,218,73,62,
		     241,24,210,44,4,20,0,248,116,49,135,100,110,130,181,169,
		     84,75,159,2,1,0,132,192,8,0,0,18,22};

  int sixsize=7;
  static int six[]={17,177,170,242,169,19,148};

  fprintf(stderr,"Low level bitpacking operations: \ttesting...");

  /* Test read/write together */
  /* Later we test against pregenerated bitstreams */
  __ogg_Bregister(0);
  __ogg_Bset(0);
  if(ogg_errno)report("couldn't alloc\n");

  fprintf(stderr,"\n\tsmall preclipped packing: ");
  cliptest(testbuffer1,test1size,one,onesize);
  fprintf(stderr,"\tpassed.");

  fprintf(stderr,"\n\tnull bit call: ");
  cliptest(testbuffer3,test3size,two,twosize);
  fprintf(stderr,"\t\t\tpassed.");

  fprintf(stderr,"\n\tlarge preclipped packing: ");
  cliptest(testbuffer2,test2size,three,threesize);
  fprintf(stderr,"\tpassed.");

  fprintf(stderr,"\n\t32 bit preclipped packing: ");
  for(i=0;i<test2size;i++)
    __ogg_Bout(large[i],32);
  __ogg_Bwflush(0,(char **)(&buffer),&bytes);
  __ogg_Bfill(0,(char *)buffer,bytes);
  for(i=0;i<test2size;i++){
    if(__ogg_Blook(32)!=large[i])report("read incorrect value!\n");
    __ogg_Badv(32);
    if(ogg_errno)report("out of data. failed!");
  }
  if(__ogg_Brflush(0)!=bytes)report("leftover bytes after read!\n");
  __ogg_free(buffer);
  fprintf(stderr,"\tpassed.");

  fprintf(stderr,"\n\tsmall unclipped packing: ");
  uncliptest(testbuffer1,test1size,7,four,foursize);
  fprintf(stderr,"\tpassed.");

  fprintf(stderr,"\n\tlarge unclipped packing: ");
  uncliptest(testbuffer2,test2size,17,five,fivesize);
  fprintf(stderr,"\tpassed.");

  fprintf(stderr,"\n\tsingle bit unclicpped packing: ");
  uncliptest(testbuffer3,test3size,1,six,sixsize);
  fprintf(stderr,"\tpassed.");

  fprintf(stderr,"\nAll lowlevel bitpacking operations pass.\n\n");
  return(0);
}  
#endif
#undef BUFFER_INCREMENT
