/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: huffman.c
  function: back-end huffman encoder
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: May 30 1996

 ********************************************************************/

/* This implementation is not public domain like the previous Huffman
   coders distributed with OggSquish.  If you want a PD Huffman coder,
   I recommend you pull the coder out of OggSquish 0.97. */

/* Encodes at most 15 bit keywords. You probably don't really want to
   use it for >8 bits, but it should be faster now that the renew
   function is no longer o(n^2); it's now linear after an O(n lg n)
   sort */

/* We need to add commit/undo to the read side of huff trees in the
   near future as well; this is trivial, I just don't have a reason to
   do it now.  It will be useful later for seekable large block
   structures that use dynamic libs; implement this as a special
   feature to a new B backend, such that the commit/undo won't be used
   when not necessary (thus saving a little memory). This can be done
   without disrupting anything already in place. */

/* Removed some dummy args in function calls that maintained
   uniformity with arithmetic; as it's uncertain that arith will ever
   be used in ogg past 0.97, I removed the dummy args. */

#include "ogg.h"
#include "memory.h"
#include "bitwise.h"
#include "ilog.h"
#include "huffman.h"

#ifdef SELFTESTS
#  include <stdio.h>
  int ogg_errno;
  int __ogg_memory;
  jmp_buf __ogg_jmp_env;
  int __ogg_jmp_env_set=0;
#else
  extern int ogg_errno;
  extern int __ogg_memory;
  extern jmp_buf __ogg_jmp_env;
  extern int __ogg_jmp_env_set;
#endif

static unsigned long mask[]=
{0x00000000,0x00000001,0x00000003,0x00000007,0x0000000f,
 0x0000001f,0x0000003f,0x0000007f,0x000000ff,0x000001ff,
 0x000003ff,0x000007ff,0x00000fff,0x00001fff,0x00003fff,
 0x00007fff,0x0000ffff,0x0001ffff,0x0003ffff,0x0007ffff,
 0x000fffff,0x001fffff,0x003fffff,0x007fffff,0x00ffffff,
 0x01ffffff,0x03ffffff,0x07ffffff,0x0fffffff,0x1fffffff,
 0x3fffffff,0x7fffffff,0xffffffff };

static size32 *h_weights;

/* Yup it's big; not everything is used at any one time. */
typedef struct huff_dict{
  int             dictsize; 
  int             encode_decode;  /* Encode/decode are assymetric */
  size16 tree;      /* point to root of the tree.  Never changes. */

  unsigned size16 *hist;     /* Histograms for sequential updating of dict */
  unsigned size16 total;

  /* Only maintained for reads and only if needed (the tree doesn't
     fit in the lookahead buffer */
  size16 *pointer0; /* Tree table to go directly to leaves */
  size16 *pointer1; /* starts after the leaves */
  size16 *parent;   /* exists for all nodes */ 

  size32 *codewords;/* for fast writes */
  size16 *lookahead;/* for fast reads  */
  int    *bits;     /* codeword legths */
  int    looksize;  /* how many bits of lookahead? */

  unsigned size16 initrenew;
  unsigned size16 currentrenew;
  unsigned size16 maxrenew;

  /* Not used for reading; for backing out changes */
  unsigned size16 *old_hist; /* commit structures, only for encode */
  unsigned size16 old_total;/* hists will pend with any write;
				the below will pend in a block with
				renewdict */
  size32 *old_codewords;
  int   *old_bits;     
  unsigned size16 old_currentrenew;
  unsigned size16 old_maxrenew;

} huff_dict;

static huff_dict **hlist=NULL;
static int hdicts=0;

/* We need to hack stable behavior wrt equal values into the qsort
   comparison */
/* Hack moved to a value frob in Hrenewdict */
#if defined(__GNUC__) || defined( WIN32 ) || defined( MACVER )
static int h_compare(const void *a,const void *b){
#else
static int h_compare(void *a,void *b){
#endif
  return(h_weights[*((int *)b)]-h_weights[*((int *)a)]);
}

#ifndef SELFSUFF
extern int OM_HUFF;
#endif

void __ogg_cleanupH(void){
  __ogg_ripfree(OM_HUFF);
  hlist=NULL;
  hdicts=0;
}

/* Only to be used for backing out dict mods during writes */
void __ogg_Hundodict(int handle){
  huff_dict *d=hlist[handle];
  memcpy(d->hist,d->old_hist,sizeof(unsigned size16)*d->dictsize);
  d->total=d->old_total;
  if(d->old_codewords){
    __ogg_free(d->codewords);
    __ogg_free(d->bits);
    d->codewords=d->old_codewords;
    d->bits=d->old_bits;
    d->old_codewords=NULL;
    d->old_bits=NULL;
    d->currentrenew=d->old_currentrenew;
    d->maxrenew=d->old_maxrenew;
  }
}

/* Only to be used for committing new dict during WRITES */
void __ogg_Hcheckpointdict(int handle){
  huff_dict *d=hlist[handle];
  memcpy(d->old_hist,d->hist,sizeof(unsigned size16)*d->dictsize);
  d->old_total=d->total;
  __ogg_free(d->old_codewords);
  __ogg_free(d->old_bits);
  d->old_codewords=NULL;
  d->old_bits=NULL;
}

/* This process generates different support structures for read/write */
/* This now uses the O(n lg n) huff generator.  Should be much faster */
static void __ogg_Hrenewdict(huff_dict *d,int max,int pendp){
  int *indexes,*stack;
  int count,l1,l2,l3,allookflag,i;
  long newrenew;
  int indexnum,stacknum,stacktop,tableptr;

  /* If we've enabled being able to back out of the last dict, we have
     to save our current state.  pendp should always be 0 for reads  */
  if(pendp)
    if(d->old_codewords==NULL){
      d->old_codewords=d->codewords;
      d->codewords=NULL;
      d->old_bits=d->bits;
      d->bits=NULL;
      d->old_currentrenew=d->currentrenew;
      d->old_maxrenew=d->maxrenew;
    }

  /* we need these, at least for generation */
  if(d->pointer0==NULL){
    d->pointer0=(size16 *)__ogg_malloc((d->dictsize-1)*sizeof(size16),
				       OM_HUFF);
    d->pointer1=(size16 *)__ogg_malloc((d->dictsize-1)*sizeof(size16),
				       OM_HUFF);
  }
  if(d->parent==NULL)
    d->parent=(size16 *)__ogg_malloc((((d->dictsize)<<1)-1)*sizeof(size16),
				     OM_HUFF);

  h_weights=(size32 *)__ogg_malloc(sizeof(size32)*((d->dictsize)<<1),
				   OM_TEMP);
  indexes=(int *)__ogg_malloc((d->dictsize<<1)*sizeof(int),OM_TEMP);
  stack=indexes+d->dictsize;

  /* Rebuild the huffman tree from histogram */
  /* We cannot have zero hist entries unless someone frobbed the struct */

  /* First, sort.  Value hack to add stable sort behavior */
  for(i=0;i<d->dictsize;i++)h_weights[i]=(d->hist[i]<<15)+i;
  for(count=0;count<d->dictsize;count++)indexes[count]=count;
  qsort(indexes,d->dictsize,sizeof(int),h_compare);

  /* Now we do the linear time tree building with an auxiliary stack */
  /* We don't need to pay too too much attention to microoptimization
     now that this loop is linear as the regen is only called
     occasionally */

  indexnum=(stacknum=(stacktop=(tableptr=d->dictsize-1)));
  while(stacktop>0){
    tableptr++;
    /* The two we want are at the bottom of index and/or stack... */

    /* Link pointers in the tree */
    /* select the two from index and stack with lowest weights.  This
       is complicated by checking to see if either list is exhausted.
       A proper implementation would avoid these checks, but I don't
       feel like esxpending thought on it right now. It's just a tad
       slower.*/

    if(indexnum<0){
      l1=stack[stacknum--];
      l2=stack[stacknum--];
    }else
      if(stacknum==stacktop){
	l1=indexes[indexnum--];
	l2=indexes[indexnum--];
      }else{
	l1=indexes[indexnum];
	l2=stack[stacknum];
	if(h_weights[l2]<h_weights[l1]){
	  l1=l2;
	  stacknum--;
	}else
	  indexnum--;

	if(stacknum==stacktop)
	  l2=indexes[indexnum--];
	else
	  if(indexnum<0)
	    l2=stack[stacknum--];
	  else{
	    l2=indexes[indexnum];
	    l3=stack[stacknum];
	    if(h_weights[l3]<h_weights[l2]){
	      l2=l3;
	      stacknum--;
	    }else
	      indexnum--;
	  }
      }

    d->parent[l1]=tableptr;
    d->parent[l2]=tableptr;
    d->pointer0[tableptr-d->dictsize]=l1;
    d->pointer1[tableptr-d->dictsize]=l2;

    /* Update our structures; indexes and histograms */
    
    h_weights[tableptr]=h_weights[l1]+h_weights[l2];
    stack[stacktop--]=tableptr;
  }

  __ogg_free(h_weights);
  __ogg_free(indexes);

  d->parent[tableptr]=-1;
  d->tree=tableptr;

  /* Now put the codewords into one place.... */

  if(d->codewords==NULL && (!d->encode_decode || d->looksize))
    d->codewords=(size32 *)__ogg_malloc(d->dictsize*sizeof(size32),
					OM_HUFF);
  if(d->bits==NULL && (!(d->encode_decode) || d->looksize))
    d->bits=(int *)__ogg_malloc(d->dictsize*sizeof(int),OM_HUFF);

  if(d->codewords){
    allookflag=-1;
    for(i=0;i<d->dictsize;i++){
      int key=i;
      int buff=0;
      int bits=0;
      while(d->parent[key]!=-1){
	if(d->pointer0[d->parent[key]-d->dictsize]==key)
	  buff<<=1;
	else
	  buff=(buff<<1)|1;
	bits++;
	key=d->parent[key];
      }
      d->codewords[i]=buff;
      d->bits[i]=bits;
      if(bits>d->looksize)allookflag=0;
    }
  }else
    allookflag=0;

  /* Generate lookahead for read if we have to */
  
  if(d->encode_decode && d->looksize){
    if(d->lookahead==NULL)
      d->lookahead=(size16 *)__ogg_malloc((1<<d->looksize)*sizeof(size16),
					  OM_HUFF);

    for(i=0;i<d->dictsize;i++){
      if(d->bits[i]>d->looksize){
	/* Too big to fit in the lookahead */
	/* Put the negation of the pointer to where in the tree to
           pick up where readahead left off */
	long j,ptr=d->tree,value=d->codewords[i];
	for(j=0;j<d->looksize;j++)
	  if((value>>j)&1)
	    ptr=d->pointer1[ptr-d->dictsize];
	  else
	    ptr=d->pointer0[ptr-d->dictsize];

	d->lookahead[(d->codewords[i]) & (mask[d->looksize])]=(size16)-ptr;
      }else{
	/* Whee. It fits. Update all possible matching entries in
           the lookahead */
	long j;
	for(j=0;j<(1<<(d->looksize));j+=(1<<(d->bits[i])))
	  d->lookahead[(d->codewords[i])|j]=i;
      }
    }
  }

  /* That should be everything.  Dealloc what we no longer need */
  if(d->encode_decode){
    /* We don't need the codewords when writing */
    __ogg_free(d->codewords);
    d->codewords=NULL;
  }
  
  /* we no longer need the parent struct in any case. */
  __ogg_free(d->parent);
  d->parent=NULL;

  if(!(d->encode_decode) || allookflag){
    /* If we're writing or the codewords fit entirely in the
       readahead, we don't need the pointer tree at all */
      __ogg_free(d->pointer0);
      __ogg_free(d->pointer1);
      d->pointer0=NULL;
      d->pointer1=NULL;
  }

  for(count=0;count<d->dictsize;count++)
    d->hist[count]=1;

  d->total=d->dictsize;
  newrenew=d->currentrenew<<1;
  d->currentrenew=(unsigned size16)(newrenew>max?max:newrenew);
  if(d->total>d->currentrenew)d->currentrenew=d->total;
}

void __ogg_Hresetdict(int handle){
  huff_dict *d=hlist[handle];
  long count;

  /* reset the histogram, renew, then checkpoint */

  d->total=d->dictsize;
  for(count=0;count<d->dictsize;count++)d->hist[count]=1;
  d->currentrenew=0;
  __ogg_Hrenewdict(d,0,0);
  d->currentrenew=d->initrenew;
  if(!(d->encode_decode))__ogg_Hcheckpointdict(handle);
}

/* Huff dicts live until the machine is reset */
int __ogg_Hcreatedict(int dictsize, int init_ref, int max_ref, int wr){
  int count;
  huff_dict *ret;

  hlist=(huff_dict **)__ogg_realloc(hlist,sizeof(huff_dict *)*(hdicts+1),
				    OM_HUFF);
  ret=hlist[hdicts]=(huff_dict *)__ogg_calloc(1,sizeof(huff_dict),OM_HUFF);

  ret->dictsize=dictsize;
  ret->tree=-1;
  switch(__ogg_memory){
  case -5:
    ret->looksize=0;
    break;
  default:
    ret->looksize=ilog(dictsize)+2+__ogg_memory;
    if(ret->looksize<0)ret->looksize=0;
    if(ret->looksize>10)ret->looksize=10;
  }
  ret->encode_decode=wr;
  ret->hist=(unsigned size16 *)__ogg_malloc(dictsize*sizeof(size16),
					    OM_HUFF);
  ret->old_hist=(unsigned size16 *)__ogg_malloc(dictsize*sizeof(size16),
						OM_HUFF);

  ret->total=dictsize;
  for(count=0;count<dictsize;count++)ret->hist[count]=1;
  ret->initrenew=init_ref;
  ret->currentrenew=0;
  __ogg_Hrenewdict(ret,0,0);
  ret->currentrenew=init_ref;
  if(max_ref>65535)max_ref=65535;
  ret->maxrenew=max_ref;
  if(!wr)__ogg_Hcheckpointdict(hdicts);
  return(hdicts++);
}

void __ogg_Hwrite(int handle, int outchar, int histp, int pendp){
/* Write a value to a bpipe while encoding it in a huffman tree */
  huff_dict *d=hlist[handle];

  __ogg_Bout(d->codewords[outchar],d->bits[outchar]);
  if(histp){
    d->hist[outchar]++;
    d->total++;
    if(d->total>=d->currentrenew)__ogg_Hrenewdict(d,d->maxrenew,pendp);
  }
}

/* Read a huffman encoded value from a bitbuffer.  First read the
   readahead; if it does not uniquely define a value, trace down the
   pointer tree the rest of the way.  After read, advance the
   bitpointer */

int __ogg_Hread(int handle,int histp){
  huff_dict *d=hlist[handle];
  long value,ptr;
  int bits;

  if(d->looksize){
    value=__ogg_Blook(d->looksize);
    value=d->lookahead[value];
    if(value<0){
      /* Lookahead didn't get it all. */
      ptr= -value;
      value=__ogg_Blook(24);
      value>>=(bits=d->looksize);
    }else{
      /* Lookahead full hit */
      __ogg_Badv(d->bits[value]);
      if(histp){
	d->hist[value]++;
	d->total++;
	if(d->total>=d->currentrenew)__ogg_Hrenewdict(d,d->maxrenew,0);
      }
      return(value);
    }
  }else{
    /* No lookahead */
    value=__ogg_Blook(24);
    ptr=d->tree;
    bits=0;
  }

  /* lookahead miss */

  while(1==1){
    if(ptr<d->dictsize)break;
    bits++;
    if(value&1)
      ptr=d->pointer1[ptr-d->dictsize];
    else
      ptr=d->pointer0[ptr-d->dictsize];
    value>>=1;
  }

  /* update histograms */

  if(histp){
    d->hist[ptr]++;
    d->total++;
    if(d->total>=d->currentrenew)__ogg_Hrenewdict(d,d->maxrenew,0);
  }

  __ogg_Badv(bits);
  return(ptr);
}

#ifdef SELFTESTS

static void report(char *in){
  fprintf(stderr,in);
  exit(1);
}

static void test(long *test,int testsize,int min, int max,int hist,int ahead,
	  char *t1,char *t2){
  /* Test with no updates */
  char *buffer;
  long bytes;
  int i;

  __ogg_memory=ahead;
  __ogg_Hcreatedict(9,min,max,0);
  __ogg_Hcreatedict(9,min,max,1);
  for(i=0;i<testsize;i++)
    __ogg_Hwrite(0,test[i],hist,0);
  __ogg_Bwflush(0,(char **)(&buffer),&bytes);

  if(t1)
    for(i=0;i<bytes;i++)if(buffer[i]!=t1[i])report("write error.  Failed!\n");

  __ogg_Bfill(0,buffer,bytes);
  for(i=0;i<testsize;i++)
    if(__ogg_Hread(1,hist)!=test[i])report("read mismatch.  Failed!\n");
  if(__ogg_Brflush(0)!=bytes)report("leftover bytes.  Failed!\n");
  __ogg_free(buffer);
  __ogg_cleanupH();

  __ogg_memory=-5;
  /* Test with no updates */
  __ogg_Hcreatedict(256,min,max,0);
  __ogg_Hcreatedict(256,min,max,1);
  for(i=0;i<testsize;i++)
    __ogg_Hwrite(0,test[i],hist,0);
  __ogg_Bwflush(0,(char **)(&buffer),&bytes);

  if(t2)
    for(i=0;i<bytes;i++)if(buffer[i]!=t2[i])report("write error.  Failed!\n");

  __ogg_Bfill(0,buffer,bytes);
  for(i=0;i<testsize;i++)
    if(__ogg_Hread(1,hist)!=test[i])report("read mismatch.  Failed!\n");
  if(__ogg_Brflush(0)!=bytes)report("leftover bytes.  Failed!\n");
  __ogg_free(buffer);
  __ogg_cleanupH();
}

static void testundo(long *test,int testsize,int min, int max,int hist,int ahead){
  /* Test with no updates */
  char *buffer,*buffer2,*buffer3;
  long bytes,bytes2,bytes3;
  int i;

  __ogg_memory=ahead;
  __ogg_Hcreatedict(9,min,max,0);
  __ogg_Hcreatedict(9,min,max,1);

  for(i=0;i<testsize;i++)
    __ogg_Hwrite(0,test[i],hist,0);
  __ogg_Bwflush(0,(char **)(&buffer),&bytes);
  __ogg_Hcheckpointdict(0);
  for(i=0;i<testsize;i++)
    __ogg_Hwrite(0,test[i],hist,0);
  __ogg_Bwflush(0,(char **)(&buffer2),&bytes2);
  __ogg_Hundodict(0);
  for(i=0;i<testsize;i++)
    __ogg_Hwrite(0,test[i],hist,0);
  __ogg_Bwflush(0,(char **)(&buffer3),&bytes3);

  __ogg_Bfill(0,buffer,bytes);
  for(i=0;i<testsize;i++)
    if(__ogg_Hread(1,hist)!=test[i])report("read mismatch.  Failed!\n");
  if(__ogg_Brflush(0)!=bytes)report("leftover bytes.  Failed!\n");
  __ogg_free(buffer);
  __ogg_free(buffer2);
  __ogg_Bfill(0,buffer3,bytes3);
  for(i=0;i<testsize;i++)
    if(__ogg_Hread(1,hist)!=test[i])report("read mismatch.  Failed!\n");
  if(__ogg_Brflush(0)!=bytes3)report("leftover bytes.  Failed!\n");
  __ogg_free(buffer3);
  __ogg_cleanupH();
}

int main(void){
  long testsize=19;
  static long test1[]=
    {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};
  static long test2[]=
    {1,1,1,1,1,1,1,1,1,1,1,1,1,1,8,8,8,8,8};
  static long test3[]=
    {0,1,2,3,4,5,5,6,7,7,8,1,8,1,8,1,8,1,8};
  static long test4[]=
    {0,0,0,0,0,0,0,0,4,8,8,1,2,3,4,5,7,1,8};
  char pre1[] ={-1,-1,-1,-1,-1,-1,-1,-1,-1,15,0};
  char pre2[] ={-128,-128,-128,-128,-128,-128,-128,-128,-128,-128,-128,
		-128,-128,-128,-128,-128,-128,-128,-128,0};
  char pre3[] ={-1,-1,-1,-1,-1,-1,-1,-37,54,0};
  char pre4[] ={-128,-128,-128,-128,-128,-128,-128,-128,-128,-128,-128,
		-128,-128,-128,16,16,16,16,16,0};
  char pre5[] ={-9,-96,-20,-76,-5,-3,126,63,0};
  char pre6[] ={0,-128,64,-64,32,-96,-96,96,-32,-32,16,-128,16,-128,16,
		-128,16,-128,16,0};
  char pre7[] ={119,119,119,119,-38,30,-108,-5,3,0};
  char pre8[] ={0,0,0,0,0,0,0,0,32,16,16,-128,64,-64,32,-96,-32,-128,16,0};
  char pre9[] ={-65,109,-37,-74,109,-37,-74,1,0};
  char pre10[]={-128,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,0};
  char pre11[]={-65,109,-37,-74,109,43,0,0};
  char pre12[]={-128,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-32,64,32,16,8,0};
  char pre13[]={119,15,-22,14,-35,110,-73,91,0};
  char pre14[]={0,0,-128,64,-64,32,0,80,48,32,-8,127,-16,-1,-32,-1,-63,-1,
		-125,3,0};
  char pre15[]={-73,109,-37,88,-10,65,101,11,0};
  char pre16[]={0,-1,-1,-1,-1,-1,-1,-1,-64,-32,-64,-1,-128,64,-64,32,96,-1,
		-63,1,0};

  fprintf(stderr,"Low level Huffman encoding routines: testing...");
  __ogg_Bregister(0);
  __ogg_Bset(0);

  fprintf(stderr,"\n\tLocked dictionary: ");
  test(test1,testsize,0,0,0,-5,pre1,pre2);
  test(test2,testsize,0,0,0,-5,pre3,pre4);
  test(test3,testsize,0,0,0,-5,pre5,pre6);
  test(test4,testsize,0,0,0,-5,pre7,pre8);
  fprintf(stderr,"\tpassed.");

  fprintf(stderr,"\n\tRolling dictionary: ");
  test(test1,testsize,3,3,1,-5,pre9,pre10);
  test(test2,testsize,3,3,1,-5,pre11,pre12);
  test(test3,testsize,3,3,1,-5,pre13,pre14);
  test(test4,testsize,3,3,1,-5,pre15,pre16);
  fprintf(stderr,"\tpassed.");

  /* Test with progressive lookahead */

  fprintf(stderr,"\n\tLookahead optimization:");
  test(test2,testsize,3,3,1,-4,NULL,NULL);
  test(test3,testsize,3,3,1,-4,NULL,NULL);
  test(test4,testsize,3,3,1,-4,NULL,NULL);

  test(test2,testsize,3,3,1,-3,NULL,NULL);
  test(test3,testsize,3,3,1,-3,NULL,NULL);
  test(test4,testsize,3,3,1,-3,NULL,NULL);

  test(test2,testsize,3,3,1,-2,NULL,NULL);
  test(test3,testsize,3,3,1,-2,NULL,NULL);
  test(test4,testsize,3,3,1,-2,NULL,NULL);

  test(test2,testsize,3,3,1,-1,NULL,NULL);
  test(test3,testsize,3,3,1,-1,NULL,NULL);
  test(test4,testsize,3,3,1,-1,NULL,NULL);

  test(test2,testsize,3,3,1,0,NULL,NULL);
  test(test3,testsize,3,3,1,0,NULL,NULL);
  test(test4,testsize,3,3,1,0,NULL,NULL);

  fprintf(stderr,"\tpassed.");
  
  /* Test commit/undo */

  fprintf(stderr,"\n\tDictionary commit/undo:");
  testundo(test1,testsize,3,3,1,-5);
  testundo(test2,testsize,3,3,1,-5);
  testundo(test3,testsize,3,3,1,-5);
  testundo(test4,testsize,3,3,1,-5);
  fprintf(stderr,"\tpassed.");

  fprintf(stderr,"\nAll low level Huffman routines: Passed.    \n");
  return(0);
}

#endif
