/********************************************************************
 *                                                                  *
 * This FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: memory.c
  function: malloc wrappers, ripcords and and mbuffer routines.
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Jun 21 1996

 ********************************************************************/

#include "ogg.h"
#include "memtypes.h"
#include "memory.h"
#include <malloc.h>

extern int ogg_errno;
extern jmp_buf __ogg_jmp_env;
extern int __ogg_jmp_env_set;

/* We practice 'ripcord' memory management in Ogg; this way, we can
   bail from any point at any time and not leak memory.  This also
   makes wholesale cleanup easier. */

/* The ripcord routines currently could upset a few platforms about
   alignment issues, but at the moment they don't so there's no reason
   to futz with them.  However, if Ogg is causing alignment errors,
   this is probably where they're beginning. */

#ifdef DEBUG
/* Nonempty ripcord blocks indicate leakage under some normal
   circumstances; report nonempty ripcords */
#  include <stdio.h>
#endif

typedef struct ripcord_block{
  void **pointers;
  long *insertlist; /* We can't embed this in the pointer list;
		       a pointer can have any value... */
  int ptop;
  int palloced;
  int pinsert;
} ripcord_block;

static ripcord_block **blocklist=NULL;
static long blocks=0;

static void longret(void){
  if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
#ifdef DEBUG
  fprintf(stderr,"\n__ogg_malloc error; out of memory?\n");
#endif
  exit(1);
}

static void check_reg_block(long block){
  /* Brain damage? */
  if(block<0){
    ogg_errno=5;
    longret();
  }

  /* Alloc new block space if necessary; make sure *we* don't leak */
  if(block>=blocks){
    ripcord_block **temp;
    long i;

    if(blocklist==NULL)
      temp=(ripcord_block **)malloc((block+1)*sizeof(ripcord_block *));
    else
      temp=(ripcord_block **)realloc(blocklist,(block+1)*
				     sizeof(ripcord_block *));
    if(temp==NULL){
      ogg_errno=1;
      longret();
    }
    blocklist=temp;
    for(i=blocks;i<=block;i++)blocklist[i]=NULL;
    blocks=block+1;
  }

  /* Allocate a new block struct if it isn't there already */
  if(blocklist[block]==NULL){
    ripcord_block *temp;

    temp=(ripcord_block *)calloc(1,sizeof(ripcord_block));
    if(temp==NULL){
      ogg_errno=1;
      longret();
    }
    blocklist[block]=temp;
  }
}

void __ogg_ripfree(long block){
  if(block>=0)
    if(block<blocks)
      if(blocklist[block]){
	int i;
	ripcord_block *b=blocklist[block];

	for(i=0;i<b->ptop;i++)
	  if(b->pointers[i]){
	    free(b->pointers[i]);
	  }
	if(b->pointers)free(b->pointers);
	if(b->insertlist)free(b->insertlist);
	free(b);
	blocklist[block]=NULL;
      }
}

/* inserts the new pointer in the specified block and initializes the
   head of the new pointer */
static void *__ogg_ripinsert(int block,void *ptr){
  long insert;
  ripcord_block *b;

  check_reg_block(block);
  b=blocklist[block];
  insert=b->pinsert;

  ((long *)ptr)[0]=block;
  ((long *)ptr)[1]=insert;

  if(insert>=b->palloced){
    void **tmp;
    long *itmp;
    /* Enlarge the ripcord lists; be very careful not to leak on
     malloc failure */
    if(b->pointers!=NULL){
      tmp=(void **)realloc(b->pointers,sizeof(void **)*(b->palloced+32));
      itmp=(long *)realloc(b->insertlist,sizeof(long *)*(b->palloced+32));
    }else{
      tmp=(void **)malloc(sizeof(void **)*32);
      itmp=(long *)malloc(sizeof(long *)*32);
    }
    if(!tmp || !itmp){
      if(tmp)free(tmp);
      if(itmp)free(itmp);
      ogg_errno=1;
      longret();
    }
    b->pointers=tmp;
    b->insertlist=itmp;
    b->palloced+=32;
  }
  if(insert==b->ptop)
    b->pinsert=++b->ptop;
  else
    b->pinsert=b->insertlist[insert];
  b->pointers[insert]=ptr;

  /* be absurd just to make certain... */
  return((void *)(((long *)ptr)+2));
}

/* remove that pointer from its associated block */

static int i_ogg_ripremove(void *ptr){
  int block=((long *)ptr)[0];
  int insert=((long *)ptr)[1];
  ripcord_block *b;

  b=blocklist[block];
  if(block>=blocks || block<0 || b==NULL){
    ogg_errno=5;
    longret();
  }
  if(b->pointers[insert]!=ptr){
    ogg_errno=5;
    longret();
  }
  b->insertlist[insert]=b->pinsert;
  b->pinsert=insert;
  b->pointers[insert]=NULL;
  ((long *)ptr)[0]=-1;
  ((long *)ptr)[1]=-1;
  return(block);
}

int __ogg_ripremove(void *ptr){
  ptr=(void *)(((long *)ptr)-2);
  return(i_ogg_ripremove(ptr));
}

/* We maintain our ripcorded blocks internally; alloc a new block
   when it is referenced by ogg_malloc */
/* MUST be freed with __ogg_free */

void *__ogg_malloc(long bytes,long block){
  void *ret;
  
  if(bytes<=0)return(NULL);
  
  ret=malloc(bytes+(sizeof(long)<<1));
  if(ret==NULL){
    ogg_errno=1;
    longret();
  }
  
  /* Insert our ripcord information in the head and update the block
     information */
  if(block==NOGROUP)
    return(ret);
  else
    return(__ogg_ripinsert(block,ret));
}

void *__ogg_calloc(long el, long size, long block){
  void *ret;
  if(el<=0 ||size<=0)return(NULL);
  ret=__ogg_malloc(el*size,block);
  memset(ret,0,el*size);
  return(ret);
}

void __ogg_free(void *ptr){
  if(ptr==NULL)return;
  ptr=(void *)(((long *)ptr)-2);
  i_ogg_ripremove(ptr);
  free(ptr);
}

void __ogg_ripmove(void *ptr,int newblock){
  if(ptr==NULL)return;
  /* restore the pointer */
  ptr=(void *)(((long *)ptr)-2);
  i_ogg_ripremove(ptr);
  __ogg_ripinsert(newblock,ptr);
}

void *__ogg_realloc(void *ptr,long size,long block){
  void *ret;
  if(ptr==NULL)
    return(__ogg_malloc(size,block));
  else{
    if(size<=0){
      __ogg_free(ptr);
      return(NULL);
    }

    /* restore the pointer */
    ptr=(void *)(((long *)ptr)-2);

    block=i_ogg_ripremove(ptr);
    ret=realloc(ptr,size+(sizeof(long)<<1));
    if(ret==NULL){
      ogg_errno=1;
      free(ptr);
      longret();
    }
    return(__ogg_ripinsert(block,ret));
  }
}

/* Utils for buffers, ogg_pipes and ogg_vars */

/* mbuffers are alloced as OM_TEMP.  If sent to a pipe, it will be
   moved to OM_PIPE.  mbuffers read from a pipe will be returned to
   OM_TEMP */

/* pass void pointers back out; the mbuffer struct is only used here */

void *__ogg_allocMB(int channels,long size){
  mbuffer *b;
  int count;

  if(channels<=0 || size<=0)return(NULL);
  b=(mbuffer *)__ogg_calloc(1,sizeof(mbuffer),OM_TEMP);
  b->ptr=__ogg_malloc(channels*sizeof(long *),OM_TEMP);

  for(count=0;count<channels;count++)
    b->ptr[count]=__ogg_calloc(size,sizeof(long),OM_TEMP);

  b->channels=channels;
  b->size=size;

  return((void *)b);
}

long *__ogg_ptrMB(void *b,int channel){
  if(b){
    if(((mbuffer *)b)->channels>channel && channel>=0)
      return(((mbuffer *)b)->ptr[channel]);
  }
  return(NULL);
}

long __ogg_sizeMB(void *b){
  if(b)
    return(((mbuffer *)b)->size);
  return(0);
}

long __ogg_chanMB(void *b){
  if(b)
    return(((mbuffer *)b)->channels);
  return(0);
}

void __ogg_freeMB(void *in){
  int count;
  mbuffer *b=(mbuffer *)in;
  if(b){
    for(count=0;count<b->channels;count++)
      __ogg_free(b->ptr[count]);
    __ogg_free(b->ptr);
    __ogg_free(b);
  }
}

void *__ogg_ncopyMB(void *d, void *i, int from, int num){
  int count,above;
  mbuffer *dest=(mbuffer *)d, *in=(mbuffer *)i;

  if(in==NULL){
    __ogg_freeMB(dest);
    return(NULL);
  }
  if(from+num>in->size)num=in->size-from;

  if(num<=0)return(NULL);

  if(dest==NULL){
    dest=(mbuffer *)(d=__ogg_allocMB(in->channels,num));
    if(dest==NULL)return(NULL);
    above=0;
  }else{
    if(dest->channels != in->channels)return(NULL);
    above=dest->size;
    for(count=0;count<in->channels;count++)
      dest->ptr[count]=__ogg_realloc(dest->ptr[count],(above+num)*sizeof(long),
        OM_TEMP);
  
    dest->size=above+num;
  }
  
  for(count=0;count<in->channels;count++)
      memcpy(dest->ptr[count]+above,in->ptr[count]+from,sizeof(long)*num);

  return(d);
}

void *__ogg_copyMB(void *dest,void *in){
  if(in)
    return(__ogg_ncopyMB(dest,in,0,((mbuffer *)in)->size));
  return(NULL);
}

void __ogg_setMB(void *i,int val){
  int count,b;
  mbuffer *in=(mbuffer *)i;
  if (in!=NULL)
    if(in->ptr)
      for(b=0;b<in->channels;b++)
	for(count=0;count<in->size;count++)
	  in->ptr[b][count]=val;
}

void __ogg_ripmoveMB(void *b,int group){
  int count;
  if(b){
    __ogg_ripmove(b,group);
    __ogg_ripmove(((mbuffer *)b)->ptr,group);
    for(count=0;count<((mbuffer *)b)->channels;count++)
      __ogg_ripmove(((mbuffer *)b)->ptr[count],group);
  }
}

/*** ogg_pipes and ogg_vars (used to be mpipes/mvars for some reason I
      can't remember) ****************************************************/

mpipe **__ogg_pipelist=NULL;
long __ogg_pipes=0;

#ifndef SELFSUFF
extern int OM_PIPE;
#endif

static mpipe *__ogg_i_createPIPE(void){
  mpipe *ret;
  ret=(mpipe *)__ogg_calloc(1,sizeof(mpipe),OM_PIPE);
  return(ret);
}

void __ogg_resetP(int pipenum){
  mpipe *p;

  p=__ogg_pipelist[pipenum];
  if(p)
    if(p->buffer){
      __ogg_freeMB((void *)(p->buffer));
      p->buffer=NULL;
    }
}

/* These are a buffer sink.  A buffer sent to a pipe should *not* be
   freed or used later.  There can be multiple sources to a sink */

/* all submitted buffers are moved to OM_PIPE */

void __ogg_appendtoP(int pipe,void *b){
  int count;
  mpipe *p;
  mbuffer *pb;

  p=__ogg_pipelist[pipe];
  if(b && ((mbuffer *)(b))->size){
    if(p){
      if(p->buffer){
	pb=p->buffer;
	for(count=0;count<((mbuffer *)(b))->channels;count++){
	  pb->ptr[count]=
	    __ogg_realloc(pb->ptr[count],(((mbuffer *)(b))->size+pb->size)*
	    sizeof(long),OM_PIPE);
	  memcpy(pb->ptr[count]+pb->size,((mbuffer *)(b))->ptr[count],
		 ((mbuffer *)(b))->size*sizeof(long));
	}
	pb->size+=((mbuffer *)(b))->size;
	__ogg_freeMB(b);
      }else{
	__ogg_ripmoveMB(b,OM_PIPE);
	p->buffer=((mbuffer *)(b));
      }
    }
  }
}

void __ogg_sendtoP(int pipe,void *b){
  __ogg_resetP(pipe);
  __ogg_appendtoP(pipe,b);
}

/* Copy operations do not remove data from the pipe. */
void *__ogg_copyfromP(int pipenum){
  mpipe *p;
  void *ret;

  p=__ogg_pipelist[pipenum];
  if(p)
    if(p->buffer){
      ret=__ogg_copyMB(NULL,((void *)(p->buffer)));
      __ogg_ripmoveMB(ret,OM_TEMP);
      return(ret);
    }
  return(NULL);
}

void *__ogg_copynfromP(int pipenum,long n){
  /* If we try to copy more bytes than are in the pipe, we will
     get a vector of the requested size padded with zeros from
     mbufferncopy */
  mpipe *p;
  void *ret;

  p=__ogg_pipelist[pipenum];
  if(p)
    if(p->buffer){
      ret=__ogg_ncopyMB(NULL,((void *)(p->buffer)),0,n);
      __ogg_ripmoveMB(ret,OM_TEMP);
      return(ret);
    }
  return(NULL);
}

/* Read operations return a buffer in addition to removing the returned data
   from the pipe */

void *__ogg_readfromP(int pipenum){
  mbuffer *ret;
  mpipe *p;

  p=__ogg_pipelist[pipenum];
  if(p)
    if(p->buffer){
      ret=p->buffer;
      p->buffer=NULL;
      __ogg_ripmoveMB(ret,OM_TEMP);
      return((void *)(ret));
    }
  return(NULL);
}

void __ogg_removenfromP(int pipenum, long n){
  int count;
  mpipe *p;

  p=__ogg_pipelist[pipenum];
  if(p)
    if(p->buffer)
      if(n<p->buffer->size){
	p->buffer->size-=n;
	for(count=0;count<p->buffer->channels;count++)
	  memmove(p->buffer->ptr[count],p->buffer->ptr[count]+n,
		  p->buffer->size*sizeof(long));
      }else __ogg_resetP(pipenum);
}

void *__ogg_readnfromP(int pipenum, long n){
  void *ret;

  ret=__ogg_copynfromP(pipenum,n);
  __ogg_removenfromP(pipenum,n);
  return(ret);
}

long __ogg_testP(int pipenum){
  mpipe *p;
  
  p=__ogg_pipelist[pipenum];
  if(p)
    if(p->buffer)
      return(p->buffer->size);
  return(0);
}

void __ogg_registerP(int pipenum){
  /* Creates and adds this pipe to the 'pipelist' if it has not
     already been done */
  int count;
  
  if(pipenum>=__ogg_pipes){
    __ogg_pipelist=(mpipe **)__ogg_realloc(__ogg_pipelist,
		 sizeof(mpipe *)*(pipenum+1),OM_PIPE);
    for(count=__ogg_pipes;count<=pipenum;count++)
      __ogg_pipelist[count]=NULL;
    __ogg_pipes=pipenum+1;
  }

  if(__ogg_pipelist[pipenum]==NULL)__ogg_pipelist[pipenum]=__ogg_i_createPIPE();
}

mvar **__ogg_varlist=NULL;
long __ogg_mvars=0;

#ifndef SELFSUFF
extern int OM_VAR;
#endif

static mvar *__ogg_i_createVAR(void){
  mvar *ret;
  ret=(mvar *)__ogg_calloc(1,sizeof(mvar),OM_VAR);
  ret->ival=-1;
  return(ret);
}

void __ogg_writetoV(int num,long val){
  mvar *m;

  m=__ogg_varlist[num];
  if(m)m->ival=val;
}

long __ogg_readfromV(int num){
  return(__ogg_varlist[num]->ival);
}

int __ogg_verifyV(int num){
  if(__ogg_mvars>=num)
    if(__ogg_varlist[num])
      return(1);
  return(0);
}

void __ogg_registerV(int num){
  /* Creates and adds this mvar to the 'varlist' if it has not
     already been done */
  int count;

  if(num==-1)return;
  if(num>=__ogg_mvars){
    __ogg_varlist=(mvar **)__ogg_realloc(__ogg_varlist,
					 sizeof(mvar *)*(num+1),OM_VAR);
    for(count=__ogg_mvars;count<=num;count++)
     __ogg_varlist[count]=NULL;
    __ogg_mvars=num+1;
  }

  if(__ogg_varlist[num]==NULL)__ogg_varlist[num]=__ogg_i_createVAR();
}

void __ogg_cleanupP(void){
  __ogg_pipelist=NULL;
  __ogg_pipes=0;
  __ogg_ripfree(OM_PIPE);
} 

void __ogg_cleanupV(void){
  __ogg_varlist=NULL;
  __ogg_mvars=0;
  __ogg_ripfree(OM_VAR);
} 

