/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: ogg.c
  function: toplevel interface file for libogg
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Jul 31 1996

 ********************************************************************/

#include <signal.h>
#include "ogg.h"
#include "memory.h"
#include "bitwise.h"
#include "huffman.h"
#include "sdefines.h"
#include "strerrs.h"
#include "buffer.h"
#include "virtual.h"
#include "parser.h"

/* Global error reporting */

int ogg_errno=0;       /* Will be reset by next function call */
char *ogg_strerr=NULL;

/* Ogg only has one reaction to errors; set flags and bail.  So, we
   can get away with using one level of longjmp to simplify error
   handling code. */

jmp_buf __ogg_jmp_env;
int __ogg_jmp_env_set=0;

/* We also want to trap SIGSEGV and SIGINT */

static void (*__ogg_old_term)(int);
static void (*__ogg_old_int)(int)=NULL;
static void (*__ogg_old_segv)(int);
static void (*__ogg_old_bus)(int);
static void (*__ogg_old_arith)(int);

static void __ogg_interrupt(int dummy){
  signal(SIGINT,__ogg_old_int);
  signal(SIGTERM,__ogg_old_term);
  ogg_errno=4;
  ogg_strerr=top_errors[ogg_errno];
  if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
  exit(1);
}

static void __ogg_arith(int dummy){
  signal(SIGFPE,__ogg_old_arith);
  ogg_errno=21;
  ogg_strerr=top_errors[ogg_errno];
  if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
  exit(1);
}

static void __ogg_segv(int dummy){
  signal(SIGSEGV,__ogg_old_segv);
  ogg_errno=3;
  ogg_strerr=top_errors[ogg_errno];
  if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
  exit(1);
}

#ifndef WIN32
static void __ogg_bus(int dummy){
  signal(SIGBUS,__ogg_old_bus);
  ogg_errno=16;
  ogg_strerr=top_errors[ogg_errno];
  if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
  exit(1);
}
#endif

static void __ogg_set_signals(void){
  __ogg_old_int=signal(SIGINT,__ogg_interrupt);
  __ogg_old_term=signal(SIGTERM,__ogg_interrupt);
  __ogg_old_segv=signal(SIGSEGV,__ogg_segv);
  __ogg_old_arith=signal(SIGFPE,__ogg_arith);
#ifndef WIN32
  __ogg_old_bus=signal(SIGBUS,__ogg_bus);
#endif
}

static void __ogg_reset_signals(void){
  __ogg_ripfree(OM_TEMP);
  if(__ogg_old_int){
    signal(SIGINT,__ogg_old_int);
    signal(SIGTERM,__ogg_old_term);
    signal(SIGSEGV,__ogg_old_segv);
    signal(SIGFPE,__ogg_old_arith);
#ifndef WIN32
    signal(SIGBUS,__ogg_old_bus);
#endif
    __ogg_old_int=NULL;
    __ogg_jmp_env_set=0;
  }
}

/* Toplevel parse requests ******************************************/

int ogg_script_ready=0;
extern int ogg_run_ready;
int ogg_run_loaded=0;

int ogg_script_err_line=0;    /* Need not be read/reset */
int ogg_script_err_statement=0;    /* Need not be read/reset */
int ogg_script_err_argument=0;     /* Need not be read/reset */
int ogg_script_err_number=0;       /* Need not be read/reset */
char *ogg_script_err_string=NULL;  /* Need not be read/reset */

int __ogg_script_complete=0;

static int script_begun=0;
static int header_begun=0;

/* This does not reset the machine, only the scripter;
   also, one can load a new script while the ogg_run_ready flag is set,
   one just can't burn it in.  The machine must be reset first */

static void __ogg_script_plunk(void){
  __ogg_CleanupParser();   /* virtual.c */
  __ogg_CleanupAssembler();/* virtual.c */
  __ogg_script_complete=0;
  script_begun=0;
  header_begun=0;
  ogg_script_ready=0;
  if(ogg_errno<100)ogg_strerr=top_errors[ogg_errno];
  __ogg_reset_signals();
}

void ogg_script_reset(void){
  ogg_errno=0;
  ogg_strerr=NULL;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  setjmp(__ogg_jmp_env);
  __ogg_set_signals();
  __ogg_script_plunk();
  ogg_script_err_string=NULL;
  ogg_script_err_line=-1;
  ogg_script_err_statement=-1;
  ogg_script_err_number=0;
}

void ogg_script_parse(char *script){
  ogg_errno=0;
  ogg_strerr=NULL;
  ogg_script_err_string=NULL;
  ogg_script_err_line=-1;
  ogg_script_err_statement=-1;
  ogg_script_err_number=0;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  if(setjmp(__ogg_jmp_env)){
    __ogg_script_plunk();
    return;
  }
  __ogg_set_signals();
  if(ogg_script_ready || header_begun){
    ogg_errno=11;
    ogg_strerr=top_errors[ogg_errno];
    __ogg_reset_signals();
    return;
  }
  if(!script_begun){
    __ogg_inputTOlines(standard_defines);
    script_begun=-1;
  }
  __ogg_inputTOlines(script);
  if(__ogg_script_complete){
    __ogg_firstpass();
    __ogg_secondpass();
    ogg_script_ready=-1;
    __ogg_CleanupParser();
    script_begun=0;
  }
  __ogg_reset_signals();
}

static void ogg_run_icleanup(void){
  __ogg_cleanupP();
  __ogg_cleanupV();
  __ogg_cleanupB();
  __ogg_cleanupH();
  __ogg_cleanupM();
  ogg_run_ready=0;
  ogg_run_loaded=0;
}

void ogg_script_use(int script){
  ogg_errno=0;
  ogg_strerr=NULL;
  ogg_script_err_string=NULL;
  ogg_script_err_line=-1;
  ogg_script_err_statement=-1;
  ogg_script_err_number=0;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  if(setjmp(__ogg_jmp_env)){
    /* Oh no!  Exception... */
    /* dealloc the engine too */
    ogg_run_icleanup();
    __ogg_script_plunk();
    return;
  }
  __ogg_set_signals();
  if(!ogg_script_ready){
    if(script_begun){
      ogg_errno=10;
      ogg_strerr=top_errors[ogg_errno];
    }else{
      ogg_errno=9;
      ogg_strerr=top_errors[ogg_errno];
    }
    __ogg_reset_signals();
    return;
  }
  if(ogg_run_loaded){
    ogg_errno=12;
    ogg_strerr=top_errors[ogg_errno];
    __ogg_reset_signals();
    return;
  }

  /* install the active interpreter bank */
  __ogg_bankswitch(script);
  /* Burn in the script */
  __ogg_makemachine();

  /* Set run ready, reset the scripter */
  ogg_run_ready=-1;
  ogg_run_loaded=-1;
  ogg_script_reset();
}

void ogg_script_makeheader(char **ret,long *bytes){
  ogg_errno=0;
  ogg_strerr=NULL;
  ogg_script_err_string=NULL;
  ogg_script_err_line=-1;
  ogg_script_err_statement=-1;
  ogg_script_err_number=0;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  if(setjmp(__ogg_jmp_env)){
    /* Oh no!  Exception... */
    __ogg_script_plunk();
    return;
  }
  __ogg_set_signals();
  if(!ogg_script_ready){
    if(script_begun){
      ogg_errno=10;
      ogg_strerr=top_errors[ogg_errno];
    }else{
      ogg_errno=9;
      ogg_strerr=top_errors[ogg_errno];
    }
    __ogg_reset_signals();
    return;
  }

  /* Pack the script */
  __ogg_packheader(ret,bytes);

  /* reset the scripter */
  ogg_script_reset();
}

long ogg_script_loadheader(char *header,long bytes){
  long bread;
  ogg_errno=0;
  ogg_strerr=NULL;
  ogg_script_err_string=NULL;
  ogg_script_err_line=-1;
  ogg_script_err_statement=-1;
  ogg_script_err_number=0;
  __ogg_jmp_env_set=-1;
  if(setjmp(__ogg_jmp_env)){
    __ogg_script_plunk();
    return(0);
  }
  __ogg_set_signals();
  if(ogg_script_ready || script_begun){
    ogg_errno=11;
    ogg_strerr=top_errors[ogg_errno];
    __ogg_reset_signals();
    return(0);
  }
  if(!header_begun)
    header_begun=-1;

  bread=__ogg_readheader(header,bytes);

  if(__ogg_script_complete){
    ogg_script_ready=-1;
    __ogg_CleanupParser();
    header_begun=0;
  }
  __ogg_reset_signals();
  return(bread);
}

/**** Engine settings stuff ******************************/

int __ogg_memory=0;
int __ogg_precision=0;

void ogg_settings_get(struct ogg_settings *s){
  ogg_errno=0;
  ogg_strerr=NULL;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  if(setjmp(__ogg_jmp_env)){
    __ogg_reset_signals();
    return;
  }
  __ogg_set_signals();
  s->ogg_memory=__ogg_memory;
  s->ogg_precision=__ogg_precision;
  __ogg_reset_signals();
}

void ogg_settings_submit(struct ogg_settings *s){
  ogg_errno=0;
  ogg_strerr=NULL;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  if(setjmp(__ogg_jmp_env)){
    __ogg_reset_signals();
    return;
  }
  __ogg_set_signals();

  if(s->ogg_memory<-5 || s->ogg_memory>0 ||
     s->ogg_precision<-5 || s->ogg_precision>0){
    ogg_errno=13;
    ogg_strerr=top_errors[ogg_errno];
    __ogg_reset_signals();
    return;
  }

  __ogg_memory=s->ogg_memory;
  __ogg_precision=s->ogg_precision;
  __ogg_reset_signals();
}

/* Buffer queries **********************************/

static void checkmachineready(void){
  __ogg_set_signals();
  if(ogg_run_ready)return;
  ogg_errno=20;
  ogg_strerr=top_errors[ogg_errno];
  if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
  exit(1);
}

static void checkmachineloaded(void){
  __ogg_set_signals();
  if(ogg_run_loaded)return;
  ogg_errno=19;
  ogg_strerr=top_errors[ogg_errno];
  if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
  exit(1);
}

int ogg_buffer_used(int group,int name){
  int ret;
  ogg_errno=0;
  ogg_strerr=NULL;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  if(setjmp(__ogg_jmp_env)){
    __ogg_reset_signals();
    return(0);
  }
  checkmachineloaded();
  ret=__ogg_buffer_used(group,name);
  __ogg_reset_signals();
  return(ret);
}

void ogg_buffer_dsp(int group,int name,struct ogg_dspinfo *s){
  ogg_errno=0;
  ogg_strerr=NULL;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  if(setjmp(__ogg_jmp_env)){
    __ogg_reset_signals();
    return;
  }
  checkmachineloaded();
  __ogg_buffer_dsp(group,name,s);
  __ogg_reset_signals();
}

void ogg_buffer_bit(int group,int name,struct ogg_bitinfo *s){
  ogg_errno=0;
  ogg_strerr=NULL;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  if(setjmp(__ogg_jmp_env)){
    __ogg_reset_signals();
    return;
  }
  checkmachineloaded();
  __ogg_buffer_bit(group,name,s);
  __ogg_reset_signals();
}

long ogg_input_query(int group,int name){
  long ret;
  ogg_errno=0;
  ogg_strerr=NULL;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  if(setjmp(__ogg_jmp_env)){
    __ogg_reset_signals();
    return(-1);
  }
  checkmachineready();
  ret=__ogg_input_query(group,name);
  __ogg_reset_signals();
  return(ret);
}

void ogg_input_reset(int group,int name){
  ogg_errno=0;
  ogg_strerr=NULL;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  if(setjmp(__ogg_jmp_env)){
    __ogg_reset_signals();
    return;
  }
  checkmachineready();
  __ogg_input_reset(group,name);
  __ogg_reset_signals();
}

void ogg_input_dsp(int group,int name, void *data, long elements, int width,
		   int endian, int sign, int scale){
  ogg_errno=0;
  ogg_strerr=NULL;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  if(setjmp(__ogg_jmp_env)){
    __ogg_reset_signals();
    return;
  }
  checkmachineready();
  __ogg_input_dsp(group,name,data,elements,width,endian,sign,scale);
  __ogg_reset_signals();
}

void ogg_output_dsp(int group,int name, void **data, long *elements, int width,
		    int endian, int sign, int scale){
  ogg_errno=0;
  ogg_strerr=NULL;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  if(setjmp(__ogg_jmp_env)){
    __ogg_reset_signals();
    return;
  }
  checkmachineloaded();
  __ogg_output_dsp(group,name,data,elements,width,endian,sign,scale);
  __ogg_reset_signals();
}

void ogg_input_bit(int group,int name, char *data, long elements){
  ogg_errno=0;
  ogg_strerr=NULL;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  if(setjmp(__ogg_jmp_env)){
    __ogg_reset_signals();
    return;
  }
  checkmachineready();
  __ogg_input_bit(group,name,data,elements);
  __ogg_reset_signals();
}

void ogg_output_bit(int group,int name, char **data, long *elements){
  ogg_errno=0;
  ogg_strerr=NULL;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  if(setjmp(__ogg_jmp_env)){
    __ogg_reset_signals();
    return;
  }
  checkmachineloaded();
  __ogg_output_bit(group,name,data,elements);
  __ogg_reset_signals();
}

/* Full runtime procs *******************************/

void ogg_run_set(int num){
  ogg_errno=0;
  ogg_strerr=NULL;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  if(setjmp(__ogg_jmp_env)){
    __ogg_reset_signals();
    return;
  }
  checkmachineloaded();
  __ogg_bankswitch(num);
  __ogg_reset_signals();
}

void ogg_run(void){
  ogg_errno=0;
  ogg_strerr=NULL;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  if(setjmp(__ogg_jmp_env)){
    if(ogg_errno<100)ogg_strerr=top_errors[ogg_errno];
    __ogg_reset_signals();
    return;
  }
  checkmachineready();
  __ogg_virtual_machine_pass();
  __ogg_reset_signals();
}

void ogg_run_reset(void){
  ogg_errno=0;
  ogg_strerr=NULL;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  if(setjmp(__ogg_jmp_env)){
    __ogg_reset_signals();
    return;
  }
  checkmachineloaded();
  ogg_run_icleanup();
  __ogg_reset_signals();
}

void ogg_run_clear(void){
  ogg_errno=0;
  ogg_strerr=NULL;
#ifndef DEBUG
  __ogg_jmp_env_set=-1;
#endif
  if(setjmp(__ogg_jmp_env)){
    
    __ogg_reset_signals();
    return;
  }
  checkmachineloaded();
  __ogg_run_clear();
  __ogg_reset_signals();
}

